import os

from autofit import conf
from autofit.core import model_mapper as mm
from autofit.core import non_linear as nl

from autolens.model.galaxy import galaxy, galaxy_model as gm
from autolens.model.profiles import light_profiles as lp
from autolens.pipeline import phase as ph
from test.integration import tools

dirpath = os.path.dirname(os.path.realpath(__file__))
conf.instance = conf.Config("{}/../../workspace/config".format(dirpath),
                            "{}/../../workspace/output/".format(dirpath))

dirpath = os.path.dirname(dirpath)
output_path = '{}/output'.format(dirpath)

pipeline_name = "test"


class TestPhaseModelMapper(object):
    def test_pairing_works(self):
        data_name = '/pair_floats'

        tools.reset_paths(data_name, pipeline_name, output_path)

        sersic = lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, intensity=1.0, effective_radius=1.3,
                                     sersic_index=3.0)

        lens_galaxy = galaxy.Galaxy(light_profile=sersic)

        tools.simulate_integration_image(data_name=data_name, pixel_scale=0.5, lens_galaxies=[lens_galaxy],
                                         source_galaxies=[], target_signal_to_noise=10.0)
        image = tools.load_image(data_name=data_name, pixel_scale=0.5)

        class MMPhase(ph.LensPlanePhase):

            def pass_priors(self, previous_results):
                self.lens_galaxies[0].sersic.intensity = self.lens_galaxies[0].sersic.axis_ratio

        phase = MMPhase(lens_galaxies=dict(lens=gm.GalaxyModel(sersic=lp.EllipticalSersic)),
                        optimizer_class=nl.MultiNest, phase_name="{}/phase1".format(pipeline_name))

        initial_total_priors = phase.variable.prior_count
        phase.make_analysis(image)

        assert phase.lens_galaxies[0].sersic.intensity == phase.lens_galaxies[0].sersic.axis_ratio
        assert initial_total_priors - 1 == phase.variable.prior_count
        assert len(phase.variable.flat_prior_model_tuples) == 1

        lines = list(
            filter(lambda line: "axis_ratio" in line or "intensity" in line, phase.variable.info.split("\n")))

        assert len(lines) == 2
        assert "lens_axis_ratio                                             UniformPrior, lower_limit = 0.2, " \
               "upper_limit = 1.0" in lines
        assert "lens_intensity                                              UniformPrior, lower_limit = 0.2, " \
               "upper_limit = 1.0" in lines

    def test_constants_work(self):
        name = "const_float"
        data_name = '/const_float'

        tools.reset_paths(data_name, name, output_path)

        sersic = lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, intensity=1.0, effective_radius=1.3,
                                     sersic_index=3.0)

        lens_galaxy = galaxy.Galaxy(light_profile=sersic)

        tools.simulate_integration_image(data_name=data_name, pixel_scale=0.5, lens_galaxies=[lens_galaxy],
                                         source_galaxies=[], target_signal_to_noise=10.0)
        image = tools.load_image(data_name=data_name, pixel_scale=0.5)

        class MMPhase(ph.LensPlanePhase):

            def pass_priors(self, previous_results):
                self.lens_galaxies[0].sersic.axis_ratio = 0.2
                self.lens_galaxies[0].sersic.phi = 90.0
                self.lens_galaxies[0].sersic.intensity = 1.0
                self.lens_galaxies[0].sersic.effective_radius = 1.3
                self.lens_galaxies[0].sersic.sersic_index = 3.0

        phase = MMPhase(lens_galaxies=[gm.GalaxyModel(sersic=lp.EllipticalSersic)],
                        optimizer_class=nl.MultiNest, phase_name="{}/phase1".format(name))

        phase.optimizer.n_live_points = 20
        phase.optimizer.sampling_efficiency = 0.8

        phase.make_analysis(image)

        sersic = phase.variable.lens_galaxies[0].sersic

        assert isinstance(sersic, mm.PriorModel)

        assert isinstance(sersic.axis_ratio, mm.Constant)
        assert isinstance(sersic.phi, mm.Constant)
        assert isinstance(sersic.intensity, mm.Constant)
        assert isinstance(sersic.effective_radius, mm.Constant)
        assert isinstance(sersic.sersic_index, mm.Constant)



import sys
import os
import re
import getdist
import numpy as np
from astropy import cosmology

sys.path.append("../autolens/")

from profiles import light_profiles, mass_profiles
import galaxy

slacs_mass = '{0:e}'.format(10**11.25)

sis = mass_profiles.SphericalIsothermalMassProfile(einstein_radius=1.03595)

lens_galaxy = galaxy.Galaxy(redshift=0.2803, mass_profiles=[sis])

source_galaxy = galaxy.Galaxy(redshift=0.9818)

galaxy.LensingPlanes(galaxies=[lens_galaxy, source_galaxy], cosmological_model=cosmology.LambdaCDM(H0=70, Om0=0.3, Ode0=0.7))

print(4.4*lens_galaxy.arcsec_per_kpc)

print('{0:e}'.format(lens_galaxy.mass_within_circles(radius=1.03595)))
print(slacs_mass)

print(lens_galaxy.mass_within_circles(radius=1.03595)/float(slacs_mass))


import os

from autofit.core import non_linear as nl
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import mass_profiles as mp
from test.integration import tools

dirpath = os.path.dirname(os.path.realpath(__file__))
dirpath = os.path.dirname(dirpath)
output_path = '{}/../output/positions'.format(dirpath)


def positions_pipeline():

    data_name = ''
    pipeline_name = "positions_phase"

    tools.reset_paths(data_name, pipeline_name, output_path)

    pipeline = make_positions_pipeline(pipeline_name=pipeline_name)

    results = pipeline.run(positions=[[[1.0, 1.0], [1.0, -1.0], [-1.0, 1.0], [-1.0, -1.0]]], pixel_scale=0.05)
    for result in results:
        print(result)


def make_positions_pipeline(pipeline_name):
    phase1 = ph.PhasePositions(lens_galaxies=[gm.GalaxyModel(sis=mp.SphericalIsothermal)],
                               optimizer_class=nl.MultiNest, phase_name="{}/phase1".format(pipeline_name))

    phase1.optimizer.n_live_points = 20
    phase1.optimizer.sampling_efficiency = 0.1

    return pl.PipelinePositions(pipeline_name, phase1)


if __name__ == "__main__":
    positions_pipeline()



import os
import shutil
from astropy import cosmology as cosmo
import pytest
import numpy as np

from autofit import conf
from autolens.data.imaging import image as im
from autolens.data.array import grids, mask as msk, scaled_array
from autolens.lensing.plotters import sensitivity_fitting_plotters
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from autolens.lensing import lensing_image as li
from autolens.model.galaxy import galaxy as g
from autolens.lensing import ray_tracing
from autolens.lensing import sensitivity_fitting

@pytest.fixture(name='general_config')
def test_general_config():
    general_config_path = "{}/../../test_files/configs/plotting/".format(os.path.dirname(os.path.realpath(__file__)))
    conf.instance.general = conf.NamedConfig(general_config_path + "general.ini")


@pytest.fixture(name='sensitivity_fitting_plotter_path')
def test_sensitivity_fitting_plotter_setup():
    galaxy_plotter_path = "{}/../../test_files/plotting/fitting/".format(os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(galaxy_plotter_path):
        shutil.rmtree(galaxy_plotter_path)

    os.mkdir(galaxy_plotter_path)

    return galaxy_plotter_path

@pytest.fixture(name='grids')
def test_grids():
    return grids.DataGrids.from_shape_and_pixel_scale(shape=(100, 100), pixel_scale=0.05, sub_grid_size=2)

@pytest.fixture(name='image')
def test_image():

    image = scaled_array.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)
    noise_map = im.NoiseMap(array=2.0*np.ones((3,3)), pixel_scale=1.0)
    psf = im.PSF(array=3.0*np.ones((1,1)), pixel_scale=1.0)

    return im.Image(array=image, pixel_scale=1.0, noise_map=noise_map, psf=psf, exposure_time_map=2.0 * np.ones((3, 3)),
                    background_sky_map=3.0*np.ones((3,3)))

@pytest.fixture(name='positions')
def test_positions():
    positions = [[[0.1, 0.1], [0.2, 0.2]], [[0.3, 0.3]]]
    return list(map(lambda position_set: np.asarray(position_set), positions))

@pytest.fixture(name='mask')
def test_mask():
    return msk.Mask.circular(shape=((3,3)), pixel_scale=0.1, radius_arcsec=0.1)

@pytest.fixture(name='lensing_image')
def test_lensing_image(image, mask):
    return li.LensingImage(image=image, mask=mask)

@pytest.fixture(name='fit')
def test_fit(lensing_image):

    lens_galaxy = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=1.0), redshift=1.0)
    lens_subhalo = g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=0.1), redshift=1.0)
    source_galaxy = g.Galaxy(light=lp.EllipticalSersic(intensity=1.0), redshift=2.0)

    tracer_normal = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                                        image_plane_grids=[lensing_image.grids],
                                                        cosmology=cosmo.Planck15)
    tracer_sensitivity = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy, lens_subhalo],
                                                             source_galaxies=[source_galaxy],
                                                             image_plane_grids=[lensing_image.grids],
                                                             cosmology=cosmo.Planck15)
    return sensitivity_fitting.SensitivityProfileFit(sensitivity_images=[lensing_image], tracer_normal=tracer_normal,
                                                     tracer_sensitive=tracer_sensitivity)


def test__fit_sub_plot__output_dependent_on_config(fit, general_config, sensitivity_fitting_plotter_path):

    sensitivity_fitting_plotters.plot_fitting_subplot(fit=fit, should_plot_mask=True,
                                                      output_path=sensitivity_fitting_plotter_path, output_format='png')
    assert os.path.isfile(path=sensitivity_fitting_plotter_path + 'sensitivity_fit.png')
    os.remove(path=sensitivity_fitting_plotter_path + 'sensitivity_fit.png')


from autofit import conf
from autofit.core import non_linear as nl
from autofit.core import model_mapper as mm
from autolens.pipeline import phase as ph
from autolens.model.galaxy import galaxy_model as gm
from autolens.data.imaging import image as im
from autolens.data.imaging.plotters import imaging_plotters
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.lensing.plotters import lensing_fitting_plotters

import os


# We finished the last tutorial on sour note. Our non-linear search failed miserably, and we were unable to infer a
# lens model which fitted our realistic datas-set well. In this tutorial, we're going to right our past wrongs and infer
# the correct model - not just once, but three times!

# First, lets get the config / simulation / regular loading out the way - we'll fit_normal the same regular as the previous
# tutorial.

#Setup the path for this run
path = '{}/../'.format(os.path.dirname(os.path.realpath(__file__)))

conf.instance = conf.Config(config_path=path+'/configs/4_dealing_with_failure', output_path=path+"output")

# Even with my custom config files - the non-linear searches will take a bit of time to run in this tutorial. Whilst you
# are waiting, I would skip ahead to the cells ahead of the phase-run cells, and sit back and think about the comments,
# there's a lot to take in in this tutorial so don't feel that you're in a rush!

# Alternatively, set these running and come back in 10 minutes or so - MultiNest resumes from the existing results on
# your hard-disk, so you can rerun things to get the results instantly!

# Another simulate regular function, for the same regular again.
def simulate():

    from autolens.data.array import grids
    from autolens.model.galaxy import galaxy as g
    from autolens.lensing import ray_tracing

    psf = im.PSF.simulate_as_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)
    image_plane_grids = grids.DataGrids.grids_for_simulation(shape=(130, 130), pixel_scale=0.1, psf_shape=(11, 11))

    lens_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.9, phi=45.0, intensity=0.04,
                                                             effective_radius=0.5, sersic_index=3.5),
                           mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=0.8))

    source_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.5, phi=90.0, intensity=0.03,
                                                       effective_radius=0.3, sersic_index=3.0))
    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                                 image_plane_grids=[image_plane_grids])

    image_simulated = im.Image.simulate(array=tracer.image_plane_image_for_simulation, pixel_scale=0.1,
                                                   exposure_time=300.0, psf=psf, background_sky_level=0.1, add_noise=True)

    return image_simulated

# Simulate the regular and set it up.
image = simulate()
imaging_plotters.plot_image_subplot(image=image)

### Approach 1 -  Prior Tuning ###

# The first approach we're going to take is we're going to give our non-linear search a helping hand. Lets think about
# our priors - what they're doing is telling the non-linear search where to look in parameter space. If we tell it to
# look in the right place (that is, *tune* our priors), it'll probaly find the best-fit_normal lens model.

# We've already seen that we can fully customize priors in AutoLens, so lets do it. I've set up a custom phase
# below, and specified a new set of priors that'll give the non-linear search a much better chance at inferring the
# correct model. I've also let you know what we're changing the priors from (as are specified by the
# 'config/priors/default' config files.)

class CustomPriorPhase(ph.LensSourcePlanePhase):

    def pass_priors(self, previous_results):

        # We've imported the 'model_mapper' module as 'mm' this time, to make the code more readable. We've also
        # called our lens model_galaxy 'lens' this time, for shorter more readable code.

        # By default, the prior on the x and y coordinates of a light / mass profile is a GaussianPrior with mean
        # 0.0" and sigma "1.0. However, visual inspection of our strong lens regular tells us that its clearly around
        # x = 0.0" and y = 0.0", so lets reduce where non-linear search looks for these parameters.

        self.lens_galaxies.lens.light.centre_0 = mm.UniformPrior(lower_limit=-0.05, upper_limit=0.05)
        self.lens_galaxies.lens.light.centre_1 = mm.UniformPrior(lower_limit=-0.05, upper_limit=0.05)
        self.lens_galaxies.lens.mass.centre_0 = mm.UniformPrior(lower_limit=-0.05, upper_limit=0.05)
        self.lens_galaxies.lens.mass.centre_1 = mm.UniformPrior(lower_limit=-0.05, upper_limit=0.05)

        # By default, the axis-ratio (ellipticity) of our lens model_galaxy's light pofile is a UniformPrior between 0.2 and
        # 1.0. However, by looking at the regular it looks fairly circular, so lets use a GaussianPrior nearer 1.0.
        self.lens_galaxies.lens.light.axis_ratio = mm.GaussianPrior(mean=0.8, sigma=0.15)

        # We'll also assume that the light profile's axis_ratio informs us of the mass-profile's axis_ratio, but
        # because this may not strictly be true (because of dark matter) we'll use a wider prior.
        self.lens_galaxies.lens.mass.axis_ratio = mm.GaussianPrior(mean=0.8, sigma=0.25)

        # By default, the orientation of the model_galaxy's light profile, phi, uses a UniformPrior between 0.0 and
        # 180.0 degrees. However, if you look really close at the regular (and maybe adjust the color-map of the plot),
        # you'll be able to notice that it is elliptical and that it is oriented around 45.0 degrees counter-clockwise
        # from the x-axis. Lets update our prior
        self.lens_galaxies.lens.light.phi = mm.GaussianPrior(mean=45.0, sigma=15.0)

        # Again, lets kind of assume that light's orientation traces that of mass.
        self.lens_galaxies.lens.mass.phi = mm.GaussianPrior(mean=45.0, sigma=30.0)

        # The effective radius of a light profile is its 'half-light' radius, the radius at which 50% of its
        # total luminosity is internal to the circle or ellipse defined within that radius. AutoLens assumes a
        # UniformPrior on this quantity between 0.0" and 4.0", but inspection of the regular (again, using a colormap
        # scaling) shows the lens's light doesn't extend anywhere near 4.0", so lets reduce it.
        self.lens_galaxies.lens.light.effective_radius = mm.GaussianPrior(mean=0.5, sigma=0.8)

        # Typically, we have some knowledge of what morphology our lens model_galaxy is. Infact, most strong lenses are
        # massive ellipticals, and anyone who studies model_galaxy morphology will tell you these galaxies have a Sersic index
        # near 4. So lets change our Sersic index from a UniformPrior between 0.8 and 8.0 to reflect this
        self.lens_galaxies.lens.light.sersic_index = mm.GaussianPrior(mean=4.0, sigma=1.0)

        # Finally, the 'ring' that the lensed source forms clearly has a radius of about 0.8". This is its Einstein
        # radius, so lets change the prior from a UniformPrior between 0.0" and 4.0".
        self.lens_galaxies.lens.mass.einstein_radius = mm.GaussianPrior(mean=0.8, sigma=0.2)

        # In this exercise, I'm not going to change any priors on the source model_galaxy. Whilst lens modeling experts can
        # look at a strong lens and often tell you roughly where the source-model_galaxy is be located (in the source-plane),
        # it is something of art form. Furthermore, the source's morphology can be pretty complex and it can become
        # its very diffcult to come up with a good source prior when this is the case.

# We can now create this custom phase and run it. Our non-linear search will start in a much higher likelihood region
# of parameter space.
custom_prior_phase = CustomPriorPhase(lens_galaxies=dict(lens=gm.GalaxyModel(light=lp.EllipticalSersic,
                                                                                     mass=mp.EllipticalIsothermal)),
                                      source_galaxies=dict(source=gm.GalaxyModel(light=lp.EllipticalExponential)),
                                      optimizer_class=nl.MultiNest,
                                      phase_name='4_custom_priors')
custom_prior_result = custom_prior_phase.run(image=image)

# Bam! We get a good model. The right model. A glorious model! We gave our non-linear search a helping hand, and it
# repaid us in spades!

# Check out the PDF in the 'output/howstolens/4_custom_priors/optimizer/chains/pdfs' folder - what degeneracies do you
# notice between parameters?
lensing_fitting_plotters.plot_fitting_subplot(fit=custom_prior_result.fit)

# Okay, so we've learnt that by tuning our priors to the lens we're fitting, we can increase our chance of inferring a
# good lens model. Before moving onto the next approach, lets think about the advantages and disadvantages of prior
# tuning:

# Advantage - We found the maximum likelihood solution in parameter space.
# Advantage - The phase took less time to run, because the non-linear search explored less of parameter space.
# Disadvantage - If we specified one prior incorrectly, the non-linear search would have began and therefore ended at
#                an incorrect solution.
# Disadvantage - Our phase was tailored to this specific strong lens. If we want to fit_normal a large sample of lenses, we'd
#                have to write a custom phase for every single one - this would take up a lot of our time!

### Approach 2 -  Reducing Complexity ###

# Our non-linear searched failed because we made the lens model more realistic and therefore more complex. Maybe we
# can make it less complex, whilst still keeping it fairly realistic? Maybe there are some assumptions we can make
# to reduce the number of lens model parameters and therefore dimensionality of non-linear parameter space?

# Well, we're scientists, so we can *always* make assumptions. Below, I'm going to create a phase that assumes that
# light-traces-mass. That is, that our light-profile's origin, axis_ratio and orientation are perfectly
# aligned with its mass. This may, or may not, be a reasonable assumption, but it'll remove 4 parameters from the lens
# model (the mass-profiles x, y, axis_ratio and phi), so its worth trying!

class LightTracesMassPhase(ph.LensSourcePlanePhase):

    def pass_priors(self, previous_results):

        # In the pass priors function, we can 'pair' any two parameters by setting them equal to one another. This
        # removes the parameter on the left-hand side of the pairing from the lens model, such that is always assumes
        # the same value as the parameter on the right-hand side.
        self.lens_galaxies.lens.mass.centre_0 = self.lens_galaxies.lens.light.centre_0

        # Now, the mass-profile's x coordinate will only use the x coordinate of the light profile. Lets do this with
        # the remaining geometric parameters of the light and mass profiles
        self.lens_galaxies.lens.mass.centre_1 = self.lens_galaxies.lens.light.centre_1
        self.lens_galaxies.lens.mass.axis_ratio = self.lens_galaxies.lens.light.axis_ratio
        self.lens_galaxies.lens.mass.phi = self.lens_galaxies.lens.light.phi

# Again, we create this phase and run it. The non-linear search has a less complex parameter space to seach, and thus
# more chance finding its highest likelihood regions. (again, the results have been precomputed for your convience).
light_traces_mass_phase = LightTracesMassPhase(lens_galaxies=dict(lens=gm.GalaxyModel(light=lp.EllipticalSersic,
                                                                                       mass=mp.EllipticalIsothermal)),
                                      source_galaxies=dict(source=gm.GalaxyModel(light=lp.EllipticalExponential)),
                                      optimizer_class=nl.MultiNest,
                                               phase_name='4_light_traces_mass')

light_traces_mass_phase_result = light_traces_mass_phase.run(image=image)
lensing_fitting_plotters.plot_fitting_subplot(fit=light_traces_mass_phase_result.fit)

# The results look pretty good. Our source model_galaxy fits the datas pretty well, and we've clearly inferred a model that
# looks similar to the one above. However, inspection of the residuals shows that the fit_normal wasn't quite as good as the
# custom-phase above.

# It turns out that when I simulated this regular, light didn't perfectly trace mass. The light-profile's axis-ratio was
# 0.9, whereas the mass-profiles was 0.8. The quality of the fit_normal has suffered as a result, and the likelihood we've
# inferred is lower.
#
# Herein lies the pitfalls of making assumptions in science - they may make your model less realistic and your results
# worse! Nevertheless, our lens model is clearly much better than it was in the previous tutorial, so making
# assumptions isn't a bad idea if you're struggling to fit_normal the datas t all.

# Again, lets consider the advantages and disadvantages of this approach:

# Advantage - By reducing parameter space's complexity, we inferred the global maximum likelihood.
# Advantage - The phase is not specific to one lens - we could run it on many strong lens regular.
# Disadvantage - Our model was less realistic, and our fit_normal suffered as a result.


### Approach 3 -  Look Harder ###

# In approaches 1 and 2, we extended our non-linear search an olive branch and generously helped it find the highest
# likelihood regions of parameter space. In approach 3 ,we're going to be stern with our non-linear search, and tell it
# 'look harder'.

# Basically, every non-linear search algorithm has a set of parameters that govern how thoroughly it searches parameter
# space. The more thoroughly it looks, the more like it is that it'll find the global maximum lens model. However,
# the search will also take longer - and we don't want it to take too long to get some results.

# Lets setup a phase, and overwrite some of the non-linear search's parameters from the defaults it assumes in the
# 'config/non_linear.ini' config file:

custom_non_linear_phase = ph.LensSourcePlanePhase(lens_galaxies=dict(lens=gm.GalaxyModel(light=lp.EllipticalSersic,
                                                                                mass=mp.EllipticalIsothermal)),
                                      source_galaxies=dict(source=gm.GalaxyModel(light=lp.EllipticalExponential)),
                                      optimizer_class=nl.MultiNest,
                                                  phase_name='4_custom_non_linear')

# The 'optimizer' below is MultiNest, the non-linear search we're using.

# When MultiNest searches non-linear parameter space, it places down a set of 'live-points', each of which corresponds
# to a particular lens model (set of parameters) and has an associted likelihood. If it guesses a new lens model with a
# higher likelihood than any of the currently active live points, this new lens model will become an active point. As a
# result, the active point with the lowest likelihood is discarded.

# The more live points MultiNest uses, the more thoroughly it will sample parameter space. Lets increase the number of
# points from the default value (50) to 100.
custom_non_linear_phase.optimizer.n_live_points = 100

# When MultiNest thinks its found a 'peak' likelihood in parameter space, it begins to converge around this peak. It
# does this by guessing lens models with similar parameters. However, this peak might not be the global maximum,
# and if MultiNest converges too quickly around a peak it won't realise this before its too late.
#
# The sampling efficiency therefore describes how quickly MultiNest converges around a peak. It assumes values between
# 0.0 and 1.0, where 1.0 corresponds to the fastest convergance but highest risk of not locating the global maximum.
# Lets reduce the sampling efficiency from 0.8 to 0.5.
custom_non_linear_phase.optimizer.sampling_efficiency = 0.5

# These are the two most important MultiNest parameters controlling how it navigates parameter space, so lets run this
# phase and see if our more detailed inspection of parameter space finds the correct lens model.
custom_non_linear_result = custom_non_linear_phase.run(image=image)

# Indeed, it does. Thus, we can always brute-force our way to a good lens model, if all else fails.
lensing_fitting_plotters.plot_fitting_subplot(fit=custom_non_linear_result.fit)

# Finally, lets list the advantages and disadvantages of this approach:

# Advantage - Its easy to setup, we just increase n_live_points or decrease sampling_efficiency.
# Advantage - It generalizes to any strong lens.
# Advantage - We didn't have to make our model less realistic.
# Disadvantage - Its expensive. Very expensive. The run-time of this phase was over 6 hours. For more complex models
#                we could be talking days or weeks (or, dare I say it, months).

# So, there we have it, we can now fit_normal strong lenses PyAutoLens. And if it fails, we know how to get it to work. I hope
# you're feeling pretty smug. You might even be thinking 'why should I bother with the rest of these tutorials, if I
# can fit_normal strong a lens already'.

# Well, my friend,  I want you to think about the last disadvantage listed above. If modeling a single lens could
# really take as long as a month, are you really willing to spend your valuable time waiting for this? I'm not, which
# is why I developed AutoLens, and in the next tutorial we'll see how we can get the best of both worlds - realistic,
# complex lens model that take mere hours to infer!

# Before doing that though, I want you to go over the advantages and disadvantages listed above again, and think
# whether we could combine these different approaches to get the best of all worlds.


import numpy as np
import pytest

from autolens.data.fitting.util import fitting_util
from autolens.data.array import mask as msk, scaled_array as sca
from autolens.model.galaxy import galaxy_data
from autolens.model.galaxy import galaxy as g, galaxy_fitting
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from test.mock.mock_galaxy import MockGalaxy


class TestGalaxyFit:

    class TestLikelihood:

        def test__1x1_image__light_profile_fits_data_perfectly__lh_is_noise_term(self):

            array = sca.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)

            noise_map = sca.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)

            mask = msk.Mask(array=np.array([[True, True, True],
                                           [True, False, True],
                                           [True, True, True]]), pixel_scale=1.0)

            g0 = MockGalaxy(value=1.0)

            g_data = galaxy_data.GalaxyDataIntensities(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            fit = galaxy_fitting.GalaxyFit(galaxy_datas=[g_data], model_galaxy=g0)
            assert fit.model_galaxy == g0
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

            g_data = galaxy_data.GalaxyDataSurfaceDensity(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            fit = galaxy_fitting.GalaxyFit(galaxy_datas=[g_data], model_galaxy=g0)
            assert fit.model_galaxy == g0
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

            g_data = galaxy_data.GalaxyDataPotential(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            fit = galaxy_fitting.GalaxyFit(galaxy_datas=[g_data], model_galaxy=g0)
            assert fit.model_galaxy == g0
            assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

            g_data_y = galaxy_data.GalaxyDataDeflectionsY(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            g_data_x = galaxy_data.GalaxyDataDeflectionsX(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            fit = galaxy_fitting.GalaxyFitDeflections(galaxy_datas=[g_data_y, g_data_x], model_galaxy=g0)
            assert fit.model_galaxy == g0
            assert fit.likelihoods[0] == -0.5 * np.log(2 * np.pi * 1.0)
            assert fit.likelihoods[1] == -0.5 * np.log(2 * np.pi * 1.0)

        def test__1x2_image__noise_not_1__all_terms_correct(self):

            array = sca.ScaledSquarePixelArray(array=5.0*np.ones((3, 4)), pixel_scale=1.0)
            array[1,2] = 4.0

            noise_map = sca.ScaledSquarePixelArray(array=2.0*np.ones((3, 4)), pixel_scale=1.0)

            mask = msk.Mask(array=np.array([[True, True, True, True],
                                           [True, False, False, True],
                                           [True, True, True, True]]), pixel_scale=1.0)

            g0 = MockGalaxy(value=1.0)

            g_data = galaxy_data.GalaxyDataIntensities(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            fit = galaxy_fitting.fit_galaxy_data_with_galaxy(galaxy_datas=[g_data], model_galaxy=g0)
            assert fit.model_galaxy == g0
            assert fit.chi_squared_term == (25.0 / 4.0)
            assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
            assert fit.likelihood == -0.5 * ((25.0 / 4.0) + 2.0*np.log(2 * np.pi * 2.0**2))

            g_data = galaxy_data.GalaxyDataSurfaceDensity(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            fit = galaxy_fitting.fit_galaxy_data_with_galaxy(galaxy_datas=[g_data], model_galaxy=g0)
            assert fit.model_galaxy == g0
            assert fit.chi_squared_term == (25.0 / 4.0)
            assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
            assert fit.likelihood == -0.5 * ((25.0 / 4.0) + 2.0*np.log(2 * np.pi * 2.0**2))

            g_data = galaxy_data.GalaxyDataPotential(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            fit = galaxy_fitting.fit_galaxy_data_with_galaxy(galaxy_datas=[g_data], model_galaxy=g0)
            assert fit.model_galaxy == g0
            assert fit.chi_squared_term == (25.0 / 4.0)
            assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
            assert fit.likelihood == -0.5 * ((25.0 / 4.0) + 2.0*np.log(2 * np.pi * 2.0**2))

            g_data_y = galaxy_data.GalaxyDataDeflectionsY(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            g_data_x = galaxy_data.GalaxyDataDeflectionsX(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            fit = galaxy_fitting.fit_galaxy_data_with_galaxy(galaxy_datas=[g_data_y, g_data_x], model_galaxy=g0)
            assert fit.model_galaxy == g0
            assert fit.chi_squared_terms[0] == (25.0 / 4.0)
            assert fit.reduced_chi_squareds[0] == (25.0 / 4.0) / 2.0
            assert fit.likelihoods[0] == -0.5 * ((25.0 / 4.0) + 2.0*np.log(2 * np.pi * 2.0**2))
            assert fit.chi_squared_terms[1] == (25.0 / 4.0)
            assert fit.reduced_chi_squareds[1] == (25.0 / 4.0) / 2.0
            assert fit.likelihoods[1] == -0.5 * ((25.0 / 4.0) + 2.0*np.log(2 * np.pi * 2.0**2))

        def test__same_as_above_but_fast_likelihood(self):

            array = sca.ScaledSquarePixelArray(array=5.0*np.ones((3, 4)), pixel_scale=1.0)
            array[1,2] = 4.0

            noise_map = sca.ScaledSquarePixelArray(array=2.0*np.ones((3, 4)), pixel_scale=1.0)

            mask = msk.Mask(array=np.array([[True, True, True, True],
                                           [True, False, False, True],
                                           [True, True, True, True]]), pixel_scale=1.0)

            g0 = MockGalaxy(value=1.0)

            g_data = galaxy_data.GalaxyDataIntensities(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            assert galaxy_fitting.fast_likelihood_from_galaxy_data_and_galaxy(galaxy_datas=[g_data], model_galaxy=g0) == \
                   -0.5 * (25.0 / 4.0 + 2.0*np.log(2 * np.pi * 2.0**2))
            assert galaxy_fitting.GalaxyFit.fast_likelihood(galaxy_datas=[g_data], galaxy=g0) == \
                   -0.5 * (25.0 / 4.0 + 2.0*np.log(2 * np.pi * 2.0**2))

            g_data = galaxy_data.GalaxyDataSurfaceDensity(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            assert galaxy_fitting.fast_likelihood_from_galaxy_data_and_galaxy(galaxy_datas=[g_data], model_galaxy=g0) == \
                   -0.5 * (25.0 / 4.0 + 2.0*np.log(2 * np.pi * 2.0**2))
            assert galaxy_fitting.GalaxyFit.fast_likelihood(galaxy_datas=[g_data], galaxy=g0) == \
                   -0.5 * (25.0 / 4.0 + 2.0*np.log(2 * np.pi * 2.0**2))

            g_data = galaxy_data.GalaxyDataPotential(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            assert galaxy_fitting.fast_likelihood_from_galaxy_data_and_galaxy(galaxy_datas=[g_data], model_galaxy=g0) == \
                   -0.5 * (25.0 / 4.0 + 2.0*np.log(2 * np.pi * 2.0**2))
            assert galaxy_fitting.GalaxyFit.fast_likelihood(galaxy_datas=[g_data], galaxy=g0) == \
                   -0.5 * (25.0 / 4.0 + 2.0*np.log(2 * np.pi * 2.0**2))

            g_data_y = galaxy_data.GalaxyDataDeflectionsY(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            g_data_x = galaxy_data.GalaxyDataDeflectionsX(array=array, noise_map=noise_map, mask=mask, sub_grid_size=1)
            assert galaxy_fitting.fast_likelihood_from_galaxy_data_and_galaxy(galaxy_datas=[g_data_y, g_data_x], model_galaxy=g0) == \
                   2.0 * -0.5 * (25.0 / 4.0 + 2.0*np.log(2 * np.pi * 2.0**2))

    class TestCompareToManual:

        def test__intensities(self):

            im = sca.ScaledSquarePixelArray(array=np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                            [0.0, 1.0, 2.0, 3.0, 0.0],
                                                            [0.0, 4.0, 5.0, 6.0, 0.0],
                                                            [0.0, 7.0, 8.0, 9.0, 0.0],
                                                            [0.0, 0.0, 0.0, 0.0, 0.0]]), pixel_scale=1.0)
            mask = msk.Mask(array=np.array([[True, True, True, True, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, True, True, True, True]]), pixel_scale=1.0)
            noise_map = 2.0 * np.ones((5, 5))

            galaxy = g.Galaxy(light=lp.SphericalSersic(centre=(1.0, 2.0), intensity=1.0))
            g_data = galaxy_data.GalaxyDataIntensities(array=im, mask=mask, noise_map=noise_map, sub_grid_size=2)
            fit = galaxy_fitting.fit_galaxy_data_with_galaxy(galaxy_datas=[g_data], model_galaxy=galaxy)

            assert fit.model_galaxy == galaxy

            model_data = galaxy.intensities_from_grid(grid=g_data.grids.sub)
            model_datas = [g_data.grids.sub.sub_data_to_regular_data(sub_array=model_data)]
            residuals = fitting_util.residuals_from_datas_and_model_datas([g_data[:]], model_datas)
            chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals, [g_data.noise_map_])

            assert g_data.grids.regular.scaled_array_from_array_1d(g_data.noise_map_) == \
                   pytest.approx(fit.noise_map, 1e-4)
            assert g_data.grids.regular.scaled_array_from_array_1d(model_datas[0]) == \
                   pytest.approx(fit.model_data, 1e-4)
            assert g_data.grids.regular.scaled_array_from_array_1d(residuals[0]) == \
                   pytest.approx(fit.residual, 1e-4)
            assert g_data.grids.regular.scaled_array_from_array_1d(chi_squareds[0]) == \
                   pytest.approx(fit.chi_squared, 1e-4)

            chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds)
            noise_terms = fitting_util.noise_terms_from_noise_maps([g_data.noise_map_])
            likelihoods = fitting_util.likelihoods_from_chi_squareds_and_noise_terms(chi_squared_terms, noise_terms)

            assert likelihoods[0] == pytest.approx(fit.likelihood, 1e-4)

            fast_likelihood = galaxy_fitting.fast_likelihood_from_galaxy_data_and_galaxy(galaxy_datas=[g_data],
                                                                                         model_galaxy=galaxy)
            assert fast_likelihood == pytest.approx(fit.likelihood)

        def test__surface_density(self):

            im = sca.ScaledSquarePixelArray(array=np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                            [0.0, 1.0, 2.0, 3.0, 0.0],
                                                            [0.0, 4.0, 5.0, 6.0, 0.0],
                                                            [0.0, 7.0, 8.0, 9.0, 0.0],
                                                            [0.0, 0.0, 0.0, 0.0, 0.0]]), pixel_scale=1.0)
            mask = msk.Mask(array=np.array([[True, True, True, True, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, True, True, True, True]]), pixel_scale=1.0)
            noise_map = 2.0 * np.ones((5, 5))

            galaxy = g.Galaxy(mass=mp.SphericalIsothermal(centre=(1.0, 2.0), einstein_radius=1.0))
            g_data = galaxy_data.GalaxyDataSurfaceDensity(array=im, mask=mask, noise_map=noise_map, sub_grid_size=2)
            fit = galaxy_fitting.fit_galaxy_data_with_galaxy(galaxy_datas=[g_data], model_galaxy=galaxy)

            assert fit.model_galaxy == galaxy

            model_data = galaxy.surface_density_from_grid(grid=g_data.grids.sub)
            model_datas = [g_data.grids.sub.sub_data_to_regular_data(sub_array=model_data)]
            residuals = fitting_util.residuals_from_datas_and_model_datas([g_data[:]], model_datas)
            chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals, [g_data.noise_map_])

            assert g_data.grids.regular.scaled_array_from_array_1d(g_data.noise_map_) == \
                   pytest.approx(fit.noise_map, 1e-4)
            assert g_data.grids.regular.scaled_array_from_array_1d(model_datas[0]) == \
                   pytest.approx(fit.model_data, 1e-4)
            assert g_data.grids.regular.scaled_array_from_array_1d(residuals[0]) == \
                   pytest.approx(fit.residual, 1e-4)
            assert g_data.grids.regular.scaled_array_from_array_1d(chi_squareds[0]) == \
                   pytest.approx(fit.chi_squared, 1e-4)

            chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds)
            noise_terms = fitting_util.noise_terms_from_noise_maps([g_data.noise_map_])
            likelihoods = fitting_util.likelihoods_from_chi_squareds_and_noise_terms(chi_squared_terms, noise_terms)

            assert likelihoods[0] == pytest.approx(fit.likelihood, 1e-4)

            fast_likelihood = galaxy_fitting.fast_likelihood_from_galaxy_data_and_galaxy(galaxy_datas=[g_data],
                                                                                         model_galaxy=galaxy)
            assert fast_likelihood == pytest.approx(fit.likelihood)
            
        def test__potential(self):

            im = sca.ScaledSquarePixelArray(array=np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                            [0.0, 1.0, 2.0, 3.0, 0.0],
                                                            [0.0, 4.0, 5.0, 6.0, 0.0],
                                                            [0.0, 7.0, 8.0, 9.0, 0.0],
                                                            [0.0, 0.0, 0.0, 0.0, 0.0]]), pixel_scale=1.0)
            mask = msk.Mask(array=np.array([[True, True, True, True, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, True, True, True, True]]), pixel_scale=1.0)
            noise_map = 2.0 * np.ones((5, 5))

            galaxy = g.Galaxy(mass=mp.SphericalIsothermal(centre=(1.0, 2.0), einstein_radius=1.0))
            g_data = galaxy_data.GalaxyDataPotential(array=im, mask=mask, noise_map=noise_map, sub_grid_size=2)
            fit = galaxy_fitting.fit_galaxy_data_with_galaxy(galaxy_datas=[g_data], model_galaxy=galaxy)

            assert fit.model_galaxy == galaxy

            model_data = galaxy.potential_from_grid(grid=g_data.grids.sub)
            model_datas = [g_data.grids.sub.sub_data_to_regular_data(sub_array=model_data)]
            residuals = fitting_util.residuals_from_datas_and_model_datas([g_data[:]], model_datas)
            chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals, [g_data.noise_map_])

            assert g_data.grids.regular.scaled_array_from_array_1d(g_data.noise_map_) == \
                   pytest.approx(fit.noise_map, 1e-4)
            assert g_data.grids.regular.scaled_array_from_array_1d(model_datas[0]) == \
                   pytest.approx(fit.model_data, 1e-4)
            assert g_data.grids.regular.scaled_array_from_array_1d(residuals[0]) == \
                   pytest.approx(fit.residual, 1e-4)
            assert g_data.grids.regular.scaled_array_from_array_1d(chi_squareds[0]) == \
                   pytest.approx(fit.chi_squared, 1e-4)

            chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds)
            noise_terms = fitting_util.noise_terms_from_noise_maps([g_data.noise_map_])
            likelihoods = fitting_util.likelihoods_from_chi_squareds_and_noise_terms(chi_squared_terms, noise_terms)

            assert likelihoods[0] == pytest.approx(fit.likelihood, 1e-4)

            fast_likelihood = galaxy_fitting.fast_likelihood_from_galaxy_data_and_galaxy(galaxy_datas=[g_data],
                                                                                         model_galaxy=galaxy)
            assert fast_likelihood == pytest.approx(fit.likelihood)
            
        
        def test__deflections(self):

            im_y = sca.ScaledSquarePixelArray(array=np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                            [0.0, 1.0, 2.0, 3.0, 0.0],
                                                            [0.0, 4.0, 5.0, 6.0, 0.0],
                                                            [0.0, 7.0, 8.0, 9.0, 0.0],
                                                            [0.0, 0.0, 0.0, 0.0, 0.0]]), pixel_scale=1.0)

            im_x = sca.ScaledSquarePixelArray(array=np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                                            [0.0, 1.0, 2.0, 3.0, 0.0],
                                                            [0.0, 4.0, 0.0, 6.0, 0.0],
                                                            [0.0, 7.0, 8.0, 9.0, 0.0],
                                                            [0.0, 0.0, 0.0, 0.0, 0.0]]), pixel_scale=1.0)

            mask = msk.Mask(array=np.array([[True, True, True, True, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, False, False, False, True],
                                            [True, True, True, True, True]]), pixel_scale=1.0)
            noise_map = 2.0 * np.ones((5, 5))

            galaxy = g.Galaxy(mass=mp.SphericalIsothermal(centre=(1.0, 2.0), einstein_radius=1.0))
            g_data_y = galaxy_data.GalaxyDataDeflectionsY(array=im_y, mask=mask, noise_map=noise_map, sub_grid_size=2)
            g_data_x = galaxy_data.GalaxyDataDeflectionsX(array=im_x, mask=mask, noise_map=noise_map, sub_grid_size=2)

            fit = galaxy_fitting.fit_galaxy_data_with_galaxy(galaxy_datas=[g_data_y, g_data_x], model_galaxy=galaxy)

            assert fit.model_galaxy == galaxy

            model_data_y = galaxy.deflections_from_grid(grid=g_data_y.grids.sub)
            model_data_y = g_data_y.grids.sub.sub_data_to_regular_data(sub_array=model_data_y[:, 0])
            model_data_x = galaxy.deflections_from_grid(grid=g_data_y.grids.sub)
            model_data_x = g_data_y.grids.sub.sub_data_to_regular_data(sub_array=model_data_x[:, 1])
            model_datas = [model_data_y, model_data_x]
            residuals = fitting_util.residuals_from_datas_and_model_datas([g_data_y[:], g_data_x[:]], model_datas)
            chi_squareds = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals, [g_data_y.noise_map_,
                                                                                          g_data_x.noise_map_])

            assert g_data_y.grids.regular.scaled_array_from_array_1d(g_data_y.noise_map_) == \
                   pytest.approx(fit.noise_maps[0], 1e-4)
            assert g_data_y.grids.regular.scaled_array_from_array_1d(model_data_y) == \
                   pytest.approx(fit.model_datas[0], 1e-4)
            assert g_data_y.grids.regular.scaled_array_from_array_1d(residuals[0]) == \
                   pytest.approx(fit.residuals[0], 1e-4)
            assert g_data_y.grids.regular.scaled_array_from_array_1d(chi_squareds[0]) == \
                   pytest.approx(fit.chi_squareds[0], 1e-4)

            assert g_data_y.grids.regular.scaled_array_from_array_1d(g_data_x.noise_map_) == \
                   pytest.approx(fit.noise_maps[1], 1e-4)
            assert g_data_y.grids.regular.scaled_array_from_array_1d(model_data_x) == \
                   pytest.approx(fit.model_datas[1], 1e-4)
            assert g_data_y.grids.regular.scaled_array_from_array_1d(residuals[1]) == \
                   pytest.approx(fit.residuals[1], 1e-4)
            assert g_data_y.grids.regular.scaled_array_from_array_1d(chi_squareds[1]) == \
                   pytest.approx(fit.chi_squareds[1], 1e-4)

            chi_squared_terms = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds)
            noise_terms = fitting_util.noise_terms_from_noise_maps([g_data_y.noise_map_, g_data_x.noise_map_])
            likelihoods = fitting_util.likelihoods_from_chi_squareds_and_noise_terms(chi_squared_terms, noise_terms)

            assert likelihoods[0] == pytest.approx(fit.likelihoods[0], 1e-4)
            assert likelihoods[1] == pytest.approx(fit.likelihoods[1], 1e-4)
            assert likelihoods[0] + likelihoods[1] == pytest.approx(fit.likelihood, 1e-4)

            fast_likelihood = galaxy_fitting.fast_likelihood_from_galaxy_data_and_galaxy(
                galaxy_datas=[g_data_y, g_data_x], model_galaxy=galaxy)
            assert fast_likelihood == pytest.approx(fit.likelihood)


from functools import wraps

import numpy as np
from astropy import constants

from autolens import exc
from autolens.data.array.util import grid_util
from autolens.data.array.util import mapping_util
from autolens.data.array import grids, scaled_array


def cosmology_check(func):
    """
    Wrap the function in a function that, if the grid is a sub-grid (grids.SubGrid), rebins the computed values to  the
    datas_-grid by taking the mean of each set of sub-gridded values.

    Parameters
    ----------
    func : (profiles, *args, **kwargs) -> Object
        A function that requires transformed coordinates
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """

        Parameters
        ----------
        self
        args
        kwargs

        Returns
        -------
            A value or coordinate in the same coordinate system as those passed in.
        """

        if self.cosmology is not None:
            return func(self, *args, *kwargs)
        else:
            return None

    return wrapper


class Plane(object):

    def __init__(self, galaxies, grids, border=None, compute_deflections=True, cosmology=None):
        """A plane represents a set of galaxies at a given redshift in a ray-tracer_normal and a the grid of datas_-plane \
        or lensed coordinates.

        From a plane, the datas_'s of its galaxies can be computed (in both the datas_-plane and source-plane). The \
        surface-density, potential and deflection angles of the galaxies can also be computed.

        Parameters
        -----------
        galaxies : [Galaxy]
            The list of lens galaxies in this plane.
        grids : mask.DataGrids
            The grids of (x,y) arc-second coordinates of this plane.
        border : mask.RegularGridBorder
            The border of the regular-grid, which is used to relocate demagnified traced regular-pixel to the \
            source-plane border.
        compute_deflections : bool
            If true, the deflection-angles of this plane's coordinates are calculated use its galaxy's mass-profiles.
        """

        self.galaxies = galaxies

        if not galaxies:
            raise exc.RayTracingException('An empty list of galaxies was supplied to Plane')

        if cosmology is not None:
            if any([redshift is None for redshift in self.galaxy_redshifts]):
                raise exc.RayTracingException('A cosmology has been supplied to ray-tracing, but a galaxy does not'
                                              'have a redshift. Either do not supply a cosmology or give the galaxy'
                                              'a redshift.')

        if any([redshift is not None for redshift in self.galaxy_redshifts]):
            if not all([galaxies[0].redshift == galaxy.redshift for galaxy in galaxies]):
                raise exc.RayTracingException('The galaxies supplied to A Plane have different redshifts or one galaxy'
                                              'does not have a redshift.')

        self.grids = grids
        self.border = border

        if compute_deflections:

            def calculate_deflections(grid):
                return sum(map(lambda galaxy: galaxy.deflections_from_grid(grid), galaxies))

            self.deflections = list(map(lambda grid : grid.apply_function(calculate_deflections), self.grids))

        else:
            self.deflections = None
        self.cosmology = cosmology

    def trace_grids_to_next_plane(self):
        """Trace this plane's grids to the next plane, using its deflection angles."""

        def minus(grid, deflections):
            return grid - deflections

        return list(map(lambda grid, deflections : grid.map_function(minus, deflections), self.grids, self.deflections))

    @property
    def total_images(self):
        return len(self.grids)

    @property
    def galaxy_redshifts(self):
        return [galaxy.redshift for galaxy in self.galaxies]

    @property
    def redshift(self):
        return self.galaxies[0].redshift

    @property
    def constant_kpc(self):
        # noinspection PyUnresolvedReferences
        return constants.c.to('kpc / s').value ** 2.0 / (4 * math.pi * constants.G.to('kpc3 / M_sun s2').value)

    @property
    @cosmology_check
    def arcsec_per_kpc_proper(self):
        return self.cosmology.arcsec_per_kpc_proper(z=self.redshift).value

    @property
    @cosmology_check
    def kpc_per_arcsec_proper(self):
        return 1.0 / self.arcsec_per_kpc_proper

    @property
    @cosmology_check
    def angular_diameter_distance_to_earth(self):
        return self.cosmology.angular_diameter_distance(self.redshift).to('kpc').value

    @property
    def has_light_profile(self):
        return any(list(map(lambda galaxy: galaxy.has_light_profile, self.galaxies)))

    @property
    def has_pixelization(self):
        return any(list(map(lambda galaxy: galaxy.has_pixelization, self.galaxies)))

    @property
    def has_regularization(self):
        return any(list(map(lambda galaxy: galaxy.has_regularization, self.galaxies)))

    @property
    def has_hyper_galaxy(self):
        return any(list(map(lambda galaxy: galaxy.has_hyper_galaxy, self.galaxies)))

    @property
    def hyper_galaxies(self):
        return [galaxy.hyper_galaxy for galaxy in self.galaxies]

    @property
    def has_padded_grids(self):
        return any(list(map(lambda grid : isinstance(grid.regular, grids.PaddedRegularGrid), self.grids)))

    @property
    def mapper(self):

        galaxies_with_pixelization = list(filter(lambda galaxy: galaxy.has_pixelization, self.galaxies))

        if len(galaxies_with_pixelization) == 0:
            return None
        if len(galaxies_with_pixelization) == 1:
            pixelization = galaxies_with_pixelization[0].pixelization
            return pixelization.mapper_from_grids_and_border(self.grids[0], self.border)
        elif len(galaxies_with_pixelization) > 1:
            raise exc.PixelizationException('The number of galaxies with pixelizations in one plane is above 1')

    @property
    def regularization(self):

        galaxies_with_regularization = list(filter(lambda galaxy: galaxy.has_regularization, self.galaxies))

        if len(galaxies_with_regularization) == 0:
            return None
        if len(galaxies_with_regularization) == 1:
            return galaxies_with_regularization[0].regularization
        elif len(galaxies_with_regularization) > 1:
            raise exc.PixelizationException('The number of galaxies with regularizations in one plane is above 1')

    @property
    def image_plane_images(self):
        return list(map(lambda _image_plane_image, grid :
                        grid.regular.scaled_array_from_array_1d(_image_plane_image), self.image_plane_images_, self.grids))

    @property
    def image_plane_images_for_simulation(self):
        if not self.has_padded_grids:
            raise exc.RayTracingException(
                'To retrieve an datas_ plane datas_ for the simulation, the grids in the tracer_normal'
                'must be padded grids')
        return list(map(lambda _image_plane_image, grid : grid.regular.map_to_2d_keep_padded(_image_plane_image),
                        self.image_plane_images_, self.grids))

    @property
    def image_plane_image(self):
        return self.image_plane_images[0]

    @property
    def image_plane_image_for_simulation(self):
        return self.image_plane_images_for_simulation[0]

    @property
    def image_plane_images_(self):
        return list(map(lambda grid :
               sum([intensities_from_grid(grid.sub, [galaxy]) for galaxy in self.galaxies]), self.grids))

    @property
    def image_plane_images_of_galaxies_(self):
        return list(map(lambda grid : [intensities_from_grid(grid.sub, [galaxy]) for galaxy in self.galaxies],
                        self.grids))

    @property
    def image_plane_blurring_images_(self):
        return list(map(lambda grid :
               sum([intensities_from_grid(grid.blurring, [galaxy]) for galaxy in self.galaxies]), self.grids))

    @property
    def plane_images(self):
        return list(map(lambda grid : plane_image_from_grid_and_galaxies(shape=grid.regular.mask.shape,
                                      grid=grid.regular, galaxies=self.galaxies), self.grids))

    @property
    def surface_density(self):
        _surface_density = sum([surface_density_from_grid(self.grids[0].sub.unlensed_grid, [galaxy]) for galaxy in
                                self.galaxies])
        return self.grids[0].regular.scaled_array_from_array_1d(_surface_density)

    @property
    def potential(self):
        _potential = sum([potential_from_grid(self.grids[0].sub.unlensed_grid, [galaxy]) for galaxy in self.galaxies])
        return self.grids[0].regular.scaled_array_from_array_1d(_potential)

    @property
    def deflections_y(self):
        return self.grids[0].regular.scaled_array_from_array_1d(self.deflections_[:, 0])

    @property
    def deflections_x(self):
        return self.grids[0].regular.scaled_array_from_array_1d(self.deflections_[:, 1])

    @property
    def deflections_(self):
        return sum([deflections_from_grid(self.grids[0].sub.unlensed_grid, [galaxy]) for galaxy in self.galaxies])

    @property
    def yticks(self):
        """Compute the yticks labels of this grid, used for plotting the y-axis ticks when visualizing an datas_-grid"""
        return np.linspace(np.amin(self.grids[0].regular[:,0]), np.amax(self.grids[0].regular[:,0]), 4)

    @property
    def xticks(self):
        """Compute the xticks labels of this grid, used for plotting the x-axis ticks when visualizing an datas_-grid"""
        return np.linspace(np.amin(self.grids[0].regular[:,1]), np.amax(self.grids[0].regular[:,1]), 4)

    def luminosities_of_galaxies_within_circles(self, radius, conversion_factor=1.0):
        """
        Compute the total luminosity of all galaxies in this plane within a circle of specified radius.

        The value returned by this integral is dimensionless, and a conversion factor can be specified to convert it \
        to a physical value (e.g. the photometric zeropoint).

        See *galaxy.light_within_circle* and *light_profiles.light_within_circle* for details
        of how this is performed.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the dimensionless luminosity within.
        """
        return list(map(lambda galaxy : galaxy.luminosity_within_circle(radius, conversion_factor),
                        self.galaxies))

    def luminosities_of_galaxies_within_ellipses(self, major_axis, conversion_factor=1.0):
        """
        Compute the total luminosity of all galaxies in this plane within a ellipse of specified major-axis.

        The value returned by this integral is dimensionless, and a conversion factor can be specified to convert it \
        to a physical value (e.g. the photometric zeropoint).

        See *galaxy.light_within_ellipse* and *light_profiles.light_within_ellipse* for details
        of how this is performed.

        Parameters
        ----------
        major_axis : float
            The major-axis of the ellipse to compute the dimensionless luminosity within.
        """
        return list(map(lambda galaxy : galaxy.luminosity_within_ellipse(major_axis, conversion_factor),
                        self.galaxies))

    def masses_of_galaxies_within_circles(self, radius, conversion_factor=1.0):
        """
        Compute the total mass of all galaxies in this plane within a circle of specified radius.

        The value returned by this integral is dimensionless, and a conversion factor can be specified to convert it \
        to a physical value (e.g. the critical surface mass density).

        See *galaxy.mass_within_circle* and *mass_profiles.mass_within_circle* for details
        of how this is performed.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the dimensionless mass within.
        """
        return list(map(lambda galaxy : galaxy.mass_within_circle(radius, conversion_factor),
                        self.galaxies))

    def masses_of_galaxies_within_ellipses(self, major_axis, conversion_factor=1.0):
        """
        Compute the total mass of all galaxies in this plane within a ellipse of specified major-axis.

        The value returned by this integral is dimensionless, and a conversion factor can be specified to convert it \
        to a physical value (e.g. the critical surface mass density).

        See *galaxy.mass_within_ellipse* and *mass_profiles.mass_within_ellipse* for details
        of how this is performed.

        Parameters
        ----------
        major_axis : float
            The major-axis of the ellipse to compute the dimensionless mass within.
        """
        return list(map(lambda galaxy : galaxy.mass_within_ellipse(major_axis, conversion_factor),
                        self.galaxies))


class PlanePositions(object):

    def __init__(self, galaxies, positions, compute_deflections=True, cosmology=None):
        """A plane represents a set of galaxies at a given redshift in a ray-tracer_normal and the positions of datas_-plane \
        coordinates which mappers close to one another in the source-plane.

        Parameters
        -----------
        galaxies : [Galaxy]
            The list of lens galaxies in this plane.
        positions : [[[]]]
            The (x,y) arc-second coordinates of datas_-plane pixels which (are expected to) mappers to the same
            location(s) in the final source-plane.
        compute_deflections : bool
            If true, the deflection-angles of this plane's coordinates are calculated use its galaxy's mass-profiles.
        """

        self.galaxies = galaxies
        self.positions = positions

        if compute_deflections:
            def calculate_deflections(positions):
                return sum(map(lambda galaxy: galaxy.deflections_from_grid(positions), galaxies))

            self.deflections = list(map(lambda positions: calculate_deflections(positions), self.positions))

        self.cosmology = cosmology

    def trace_to_next_plane(self):
        """Trace the positions to the next plane."""
        return list(map(lambda positions, deflections: np.subtract(positions, deflections),
                        self.positions, self.deflections))


class PlaneImage(scaled_array.ScaledRectangularPixelArray):

    def __init__(self, array, pixel_scales, grid, origin=(0.0, 0.0)):
        self.grid = grid
        super(PlaneImage, self).__init__(array=array, pixel_scales=pixel_scales, origin=origin)


def sub_to_image_grid(func):
    """
    Wrap the function in a function that, if the grid is a sub-grid (grids.SubGrid), rebins the computed values to
    the datas_-grid by taking the mean of each set of sub-gridded values.

    Parameters
    ----------
    func : (profiles, *args, **kwargs) -> Object
        A function that requires transformed coordinates
    """

    @wraps(func)
    def wrapper(grid, galaxies, *args, **kwargs):
        """

        Parameters
        ----------
        grid : ndarray
            PlaneCoordinates in either cartesian or profiles coordinate system
        args
        kwargs

        Returns
        -------
            A value or coordinate in the same coordinate system as those passed in.
        """

        result = func(grid, galaxies, *args, *kwargs)

        if isinstance(grid, grids.SubGrid):
            return grid.sub_data_to_regular_data(result)
        else:
            return result

    return wrapper


@sub_to_image_grid
def intensities_from_grid(grid, galaxies):
    return sum(map(lambda g: g.intensities_from_grid(grid), galaxies))


@sub_to_image_grid
def surface_density_from_grid(grid, galaxies):
    return sum(map(lambda g: g.surface_density_from_grid(grid), galaxies))


@sub_to_image_grid
def potential_from_grid(grid, galaxies):
    return sum(map(lambda g: g.potential_from_grid(grid), galaxies))


# TODO : There will be a much cleaner way to apply sub datas to surface_density to the array wihtout the need for a
# transpose

def deflections_from_grid(grid, galaxies):
    deflections = sum(map(lambda galaxy: galaxy.deflections_from_grid(grid), galaxies))
    if isinstance(grid, grids.SubGrid):
        return np.asarray([grid.sub_data_to_regular_data(deflections[:, 0]), grid.sub_data_to_regular_data(deflections[:, 1])]).T
    return sum(map(lambda galaxy: galaxy.deflections_from_grid(grid), galaxies))


def deflections_from_sub_grid(sub_grid, galaxies):
    return sum(map(lambda galaxy: galaxy.deflections_from_grid(sub_grid), galaxies))


def deflections_from_grid_collection(grid_collection, galaxies):
    return grid_collection.apply_function(lambda grid: deflections_from_sub_grid(grid, galaxies))


def plane_image_from_grid_and_galaxies(shape, grid, galaxies, buffer=1.0e-2):

    y_min = np.min(grid[:, 0]) - buffer
    y_max = np.max(grid[:, 0]) + buffer
    x_min = np.min(grid[:, 1]) - buffer
    x_max = np.max(grid[:, 1]) + buffer

    pixel_scales = (float((y_max - y_min) / shape[0]), float((x_max - x_min) / shape[1]))
    origin = ((y_max + y_min) / 2.0, (x_max + x_min) / 2.0)

    uniform_grid = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(mask=np.full(shape=shape,
                                                                                                   fill_value=False),
                                                                                      pixel_scales=pixel_scales,
                                                                                      origin=origin)

    image_1d = sum([intensities_from_grid(uniform_grid, [galaxy]) for galaxy in galaxies])

    image_2d = mapping_util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(array_1d=image_1d, shape=shape)

    im = PlaneImage(array=image_2d, pixel_scales=pixel_scales, grid=grid, origin=origin)

    return PlaneImage(array=image_2d, pixel_scales=pixel_scales, grid=grid, origin=origin)


def traced_collection_for_deflections(grids, deflections):
    def subtract_scaled_deflections(grid, scaled_deflection):
        return np.subtract(grid, scaled_deflection)

    result = grids.map_function(subtract_scaled_deflections, deflections)

    return result



import os
import shutil

import numpy as np
import pytest

from autolens.data.imaging import image as im
from autolens.data.array.util import array_util
from autolens.data.array import grids, mask as msk
from autolens.lensing import lensing_fitting
from autolens.model.galaxy import galaxy as g
from autolens.lensing import lensing_image as li
from autolens.lensing import ray_tracing
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp


def test__simulate_lensed_source_and_fit__no_psf_blurring__chi_squared_is_0__noise_term_correct():

    psf = im.PSF(array=np.array([[0.0, 0.0, 0.0],
                                 [0.0, 1.0, 0.0],
                                 [0.0, 0.0, 0.0]]), pixel_scale=1.0)

    imaging_grids = grids.DataGrids.grids_for_simulation(shape=(11, 11), pixel_scale=0.2, psf_shape=psf.shape)

    lens_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.1, 0.1), intensity=0.1),
                           mass=mp.EllipticalIsothermal(centre=(0.1, 0.1), einstein_radius=1.8))
    source_galaxy = g.Galaxy(light=lp.EllipticalExponential(centre=(0.1, 0.1), intensity=0.5))
    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                                 image_plane_grids=[imaging_grids])

    image_simulated = im.Image.simulate(array=tracer.image_plane_image_for_simulation, pixel_scale=0.2,
                                                   exposure_time=300.0, psf=psf, background_sky_level=None,
                                                   add_noise=False)
    image_simulated.noise_map = np.ones(image_simulated.shape)

    path = "{}/datas".format(
        os.path.dirname(os.path.realpath(__file__)))  # Setup path so we can output the simulated datas.

    try:
        shutil.rmtree(path)
    except FileNotFoundError:
        pass

    if os.path.exists(path) == False:
        os.makedirs(path)

    array_util.numpy_array_to_fits(array=image_simulated, path=path + '/data.fits')
    array_util.numpy_array_to_fits(array=image_simulated.noise_map, path=path + '/noise_map_.fits')
    array_util.numpy_array_to_fits(array=psf, path=path + '/psf.fits')

    image = im.load_imaging_from_fits(image_path=path + '/data.fits',
                                      noise_map_path=path + '/noise_map_.fits',
                                      psf_path=path + '/psf.fits', pixel_scale=0.2)

    mask = msk.Mask.circular(shape=image.shape, pixel_scale=0.2, radius_arcsec=0.8)

    lensing_image = li.LensingImage(image=image, mask=mask, sub_grid_size=1)

    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                                 image_plane_grids=[lensing_image.grids])

    fitter = lensing_fitting.LensingProfileFit(lensing_images=[lensing_image], tracer=tracer)

    assert fitter.chi_squared_term == 0.0


def test__simulate_lensed_source_and_fit__include_psf_blurring__chi_squared_is_0__noise_term_correct():

    psf = im.PSF.simulate_as_gaussian(shape=(3, 3), pixel_scale=1.0, sigma=0.75)

    imaging_grids = grids.DataGrids.grids_for_simulation(shape=(11, 11), pixel_scale=0.2, psf_shape=psf.shape)

    lens_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.1, 0.1), intensity=0.1),
                           mass=mp.EllipticalIsothermal(centre=(0.1, 0.1), einstein_radius=1.8))
    source_galaxy = g.Galaxy(light=lp.EllipticalExponential(centre=(0.1, 0.1), intensity=0.5))
    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                                 image_plane_grids=[imaging_grids])

    image_simulated = im.Image.simulate(array=tracer.image_plane_image_for_simulation, pixel_scale=0.2,
                                                   exposure_time=300.0, psf=psf, background_sky_level=None,
                                                   add_noise=False)
    image_simulated.noise_map = np.ones(image_simulated.shape)

    path = "{}/datas".format(
        os.path.dirname(os.path.realpath(__file__)))  # Setup path so we can output the simulated datas.

    try:
        shutil.rmtree(path)
    except FileNotFoundError:
        pass

    if os.path.exists(path) == False:
        os.makedirs(path)

    array_util.numpy_array_to_fits(array=image_simulated, path=path + '/data.fits')
    array_util.numpy_array_to_fits(array=image_simulated.noise_map, path=path + '/noise_map_.fits')
    array_util.numpy_array_to_fits(array=psf, path=path + '/psf.fits')

    image = im.load_imaging_from_fits(image_path=path + '/data.fits',
                                      noise_map_path=path + '/noise_map_.fits',
                                      psf_path=path + '/psf.fits', pixel_scale=0.2)

    mask = msk.Mask.circular(shape=image.shape, pixel_scale=0.2, radius_arcsec=0.8)

    lensing_image = li.LensingImage(image=image, mask=mask, sub_grid_size=1)

    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                                 image_plane_grids=[lensing_image.grids])

    fitter = lensing_fitting.LensingProfileFit(lensing_images=[lensing_image], tracer=tracer)

    assert fitter.chi_squared_term == pytest.approx(0.0, 1e-4)



import os
import shutil

import pytest
import numpy as np

from autofit import conf
from autolens.data.imaging import image as im
from autolens.data.array import mask as msk, scaled_array
from autolens.data.imaging.plotters import imaging_plotters


@pytest.fixture(name='general_config')
def test_general_config():
    general_config_path = "{}/../../../test_files/configs/plotting/".format(os.path.dirname(os.path.realpath(__file__)))
    conf.instance.general = conf.NamedConfig(general_config_path+"general.ini")

@pytest.fixture(name='imaging_plotter_path')
def test_imaging_plotter_setup():
    imaging_plotter_path = "{}/../../../test_files/plotting/imaging/".format(os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(imaging_plotter_path):
        shutil.rmtree(imaging_plotter_path)

    os.mkdir(imaging_plotter_path)

    return imaging_plotter_path

@pytest.fixture(name='image')
def test_image():

    image = scaled_array.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)
    noise_map = im.NoiseMap(array=2.0*np.ones((3,3)), pixel_scale=1.0)
    psf = im.PSF(array=3.0*np.ones((3,3)), pixel_scale=1.0)

    return im.Image(array=image, pixel_scale=1.0, noise_map=noise_map, psf=psf)

@pytest.fixture(name='positions')
def test_positions():
    positions = [[[0.1, 0.1], [0.2, 0.2]], [[0.3, 0.3]]]
    return list(map(lambda position_set: np.asarray(position_set), positions))

@pytest.fixture(name='mask')
def test_mask():
    return msk.Mask.circular(shape=((3,3)), pixel_scale=0.1, radius_arcsec=0.1)


def test__image_sub_plot_output_dependent_on_config(image, general_config, imaging_plotter_path):

    imaging_plotters.plot_image_subplot(image=image, output_path=imaging_plotter_path, output_format='png')

    assert os.path.isfile(path=imaging_plotter_path+'regular.png')
    os.remove(path=imaging_plotter_path+'regular.png')

def test__image_individuals__output_dependent_on_config(image, general_config, imaging_plotter_path):

    imaging_plotters.plot_image_individual(image=image, output_path=imaging_plotter_path, output_format='png')

    assert os.path.isfile(path=imaging_plotter_path+'observed_image.png')
    os.remove(path=imaging_plotter_path+'observed_image.png')

    assert not os.path.isfile(path=imaging_plotter_path+'noise_map_.png')

    assert os.path.isfile(path=imaging_plotter_path+'psf.png')
    os.remove(path=imaging_plotter_path+'psf.png')

    assert not os.path.isfile(path=imaging_plotter_path+'signal_to_noise_map.png')

def test__image_is_output(image, positions, mask, imaging_plotter_path):

    imaging_plotters.plot_image(image=image, positions=positions, mask=mask, output_path=imaging_plotter_path,
                                output_format='png')
    assert os.path.isfile(path=imaging_plotter_path+'observed_image.png')
    os.remove(path=imaging_plotter_path+'observed_image.png')

def test__noise_map_is_output(image, mask, imaging_plotter_path):

    imaging_plotters.plot_noise_map(image=image, mask=mask, output_path=imaging_plotter_path, output_format='png')
    assert os.path.isfile(path=imaging_plotter_path+'noise_map_.png')
    os.remove(path=imaging_plotter_path+'noise_map_.png')

def test__psf_is_output(image, imaging_plotter_path):

    imaging_plotters.plot_psf(image=image, output_path=imaging_plotter_path, output_format='png')
    assert os.path.isfile(path=imaging_plotter_path+'psf.png')
    os.remove(path=imaging_plotter_path+'psf.png')

def test__signal_to_noise_map_is_output(image, mask, imaging_plotter_path):

    imaging_plotters.plot_signal_to_noise_map(image=image, mask=mask, output_path=imaging_plotter_path,
                                              output_format='png')
    assert os.path.isfile(path=imaging_plotter_path+'signal_to_noise_map.png')
    os.remove(path=imaging_plotter_path+'signal_to_noise_map.png')


from autolens import exc
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np
import itertools

from autolens.data.array.plotters import plotter_util


def plot_array(array, origin=None, mask=None, should_plot_border=False, positions=None, grid=None, as_subplot=False,
               units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
               cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
               cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
               title='Array', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
               mask_pointsize=10, border_pointsize=2, position_pointsize=30, grid_pointsize=1,
               xticks_manual=None, yticks_manual=None,
               output_path=None, output_format='show', output_filename='array'):

    plot_figure(array=array, as_subplot=as_subplot, units=units, kpc_per_arcsec=kpc_per_arcsec,
                figsize=figsize, aspect=aspect, cmap=cmap, norm=norm,
                norm_max=norm_max, norm_min=norm_min, linthresh=linthresh, linscale=linscale,
                xticks_manual=xticks_manual, yticks_manual=yticks_manual)

    plotter_util.set_title(title=title, titlesize=titlesize)
    set_xy_labels_and_ticksize(units=units, kpc_per_arcsec=kpc_per_arcsec, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                               xyticksize=xyticksize)

    set_colorbar(cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad)
    plot_origin(array=array, origin=origin, units=units, kpc_per_arcsec=kpc_per_arcsec)
    plot_mask(mask=mask, units=units, kpc_per_arcsec=kpc_per_arcsec, pointsize=mask_pointsize)
    plot_border(mask=mask, should_plot_border=should_plot_border, units=units, kpc_per_arcsec=kpc_per_arcsec,
                pointsize=border_pointsize)
    plot_points(points_arc_seconds=positions, array=array, units=units, kpc_per_arcsec=kpc_per_arcsec,
                pointsize=position_pointsize)
    plot_grid(grid_arc_seconds=grid, array=array, units=units, kpc_per_arcsec=kpc_per_arcsec, pointsize=grid_pointsize)
    plotter_util.output_figure(array, as_subplot=as_subplot, output_path=output_path, output_filename=output_filename,
                               output_format=output_format)
    plotter_util.close_figure(as_subplot=as_subplot)

def plot_figure(array, as_subplot, units, kpc_per_arcsec, figsize, aspect, cmap, norm, norm_max, norm_min,
                linthresh, linscale, xticks_manual, yticks_manual):

    plotter_util.setup_figure(figsize=figsize, as_subplot=as_subplot)

    norm_min, norm_max = get_normalization_min_max(array=array, norm_min=norm_min, norm_max=norm_max)
    norm_scale = get_normalization_scale(norm=norm, norm_min=norm_min, norm_max=norm_max,
                                         linthresh=linthresh, linscale=linscale)

    extent = get_extent(array=array, units=units, kpc_per_arcsec=kpc_per_arcsec,
                        xticks_manual=xticks_manual, yticks_manual=yticks_manual)

    plt.imshow(array, aspect=aspect, cmap=cmap, norm=norm_scale, extent=extent)

def get_extent(array, units, kpc_per_arcsec, xticks_manual, yticks_manual):

    if xticks_manual is not None and yticks_manual is not None:
        return np.asarray([xticks_manual[0], xticks_manual[3], yticks_manual[0], yticks_manual[3]])

    if units is 'pixels':
        return np.asarray([0, array.shape[1], 0, array.shape[0]])
    elif units is 'arcsec' or kpc_per_arcsec is None:
        return np.asarray([array.arc_second_minima[1], array.arc_second_maxima[1],
                           array.arc_second_minima[0], array.arc_second_maxima[0]])
    elif units is 'kpc':
        return list(map(lambda tick : tick*kpc_per_arcsec,
                        np.asarray([array.arc_second_minima[1], array.arc_second_maxima[1],
                                    array.arc_second_minima[0], array.arc_second_maxima[0]])))

def get_normalization_min_max(array, norm_min, norm_max):

    if norm_min is None:
        norm_min = array.min()
    if norm_max is None:
        norm_max = array.max()

    return norm_min, norm_max

def get_normalization_scale(norm, norm_min, norm_max, linthresh, linscale):
    if norm is 'linear':
        return colors.Normalize(vmin=norm_min, vmax=norm_max)
    elif norm is 'log':
        if norm_min == 0.0:
            norm_min = 1.e-4
        return colors.LogNorm(vmin=norm_min, vmax=norm_max)
    elif norm is 'symmetric_log':
        return colors.SymLogNorm(linthresh=linthresh, linscale=linscale, vmin=norm_min, vmax=norm_max)
    else:
        raise exc.PlottingException('The normalization (norm) supplied to the plotter is not a valid string (must be '
                                     'linear | log | symmetric_log')

def set_axis_limits(array, units, kpc_per_arcsec):

    if units is 'arcsec' or kpc_per_arcsec is None:

        plt.axis(axis_limits=np.asarray([array.arc_second_minima[1],
                                         array.arc_second_maxima[1],
                                         array.arc_second_minima[0],
                                         array.arc_second_maxima[0]]))

    elif units is 'kpc':

        plt.axis(axis_limits=np.asarray([array.arc_second_minima[1] * kpc_per_arcsec,
                                         array.arc_second_maxima[1] * kpc_per_arcsec,
                                         array.arc_second_minima[0] * kpc_per_arcsec,
                                         array.arc_second_maxima[0] * kpc_per_arcsec]))

def set_xy_labels_and_ticksize(units, kpc_per_arcsec, xlabelsize, ylabelsize, xyticksize):

    if units is 'pixels':

        plt.xlabel('x (pixels)', fontsize=xlabelsize)
        plt.ylabel('y (pixels)', fontsize=ylabelsize)

    elif units is 'arcsec' or kpc_per_arcsec is None:

        plt.xlabel('x (arcsec)', fontsize=xlabelsize)
        plt.ylabel('y (arcsec)', fontsize=ylabelsize)

    elif units is 'kpc':

        plt.xlabel('x (kpc)', fontsize=xlabelsize)
        plt.ylabel('y (kpc)', fontsize=ylabelsize)

    else:
        raise exc.PlottingException('The units supplied to the plotted are not a valid string (must be pixels | '
                                     'arcsec | kpc)')

    plt.tick_params(labelsize=xyticksize)

def set_colorbar(cb_ticksize, cb_fraction, cb_pad):

    cb = plt.colorbar(fraction=cb_fraction, pad=cb_pad)
    cb.ax.tick_params(labelsize=cb_ticksize)

def convert_grid_units(array, grid_arc_seconds, units, kpc_per_arcsec):

    if units is 'pixels':
        return array.grid_arc_seconds_to_grid_pixels(grid_arc_seconds=grid_arc_seconds)
    elif units is 'arcsec' or kpc_per_arcsec is None:
        return grid_arc_seconds
    elif units is 'kpc':
        return grid_arc_seconds * kpc_per_arcsec

def plot_origin(array, origin, units, kpc_per_arcsec):

    if origin is not None:

        origin_grid = np.asarray(origin)
        origin_units = convert_grid_units(array=array, grid_arc_seconds=origin_grid, units=units,
                                          kpc_per_arcsec=kpc_per_arcsec)
        plt.scatter(y=origin_units[0], x=origin_units[1], s=80, c='k', marker='x')

def plot_mask(mask, units, kpc_per_arcsec, pointsize):

    if mask is not None:

        plt.gca()
        edge_pixels = mask.masked_grid_index_to_pixel[mask.edge_pixels]
        edge_arc_seconds = mask.grid_pixels_to_grid_arc_seconds(grid_pixels=edge_pixels)
        edge_units = convert_grid_units(array=mask, grid_arc_seconds=edge_arc_seconds, units=units,
                                          kpc_per_arcsec=kpc_per_arcsec)

        plt.scatter(y=edge_units[:,0], x=edge_units[:,1], s=pointsize, c='k')

def plot_border(mask, should_plot_border, units, kpc_per_arcsec, pointsize):

    if should_plot_border and mask is not None:

        plt.gca()
        border_pixels = mask.masked_grid_index_to_pixel[mask.border_pixels]
        border_arc_seconds = mask.grid_pixels_to_grid_arc_seconds(grid_pixels=border_pixels)
        border_units = convert_grid_units(array=mask, grid_arc_seconds=border_arc_seconds, units=units,
                                          kpc_per_arcsec=kpc_per_arcsec)

        plt.scatter(y=border_units[:,0], x=border_units[:,1], s=pointsize, c='y')

def plot_points(points_arc_seconds, array, units, kpc_per_arcsec, pointsize):

    if points_arc_seconds is not None:
        points_arc_seconds = list(map(lambda position_set: np.asarray(position_set), points_arc_seconds))
        point_colors = itertools.cycle(["m", "y", "r", "w", "c", "b", "g", "k"])
        for point_set_arc_seconds in points_arc_seconds:
            point_set_units = convert_grid_units(array=array, grid_arc_seconds=point_set_arc_seconds, units=units,
                                                 kpc_per_arcsec=kpc_per_arcsec)
            plt.scatter(y=point_set_units[:,0], x=point_set_units[:,1], color=next(point_colors), s=pointsize)

def plot_grid(grid_arc_seconds, array, units, kpc_per_arcsec, pointsize):

    if grid_arc_seconds is not None:
        grid_units = convert_grid_units(grid_arc_seconds=grid_arc_seconds, array=array, units=units,
                                        kpc_per_arcsec=kpc_per_arcsec)

        plt.scatter(y=np.asarray(grid_units[:, 0]), x=np.asarray(grid_units[:, 1]), s=pointsize, c='k')


import numpy as np
import numba

@numba.jit(nopython=True, parallel=True)
def constant_regularization_matrix_from_pixel_neighbors(coefficients, pixel_neighbors, pixel_neighbors_size):
    """From the pixel-neighbors, setup the regularization matrix using the constant regularization scheme.

    Parameters
    ----------
    pixel_neighbors : [[]]
        A list of the neighbors of each pixel.
    """

    pixels = len(pixel_neighbors)

    regularization_matrix = np.zeros(shape=(pixels, pixels))

    regularization_coefficient = coefficients[0] ** 2.0

    for i in range(pixels):
        regularization_matrix[i, i] += 1e-8
        for j in range(pixel_neighbors_size[i]):
            neighbor_index = pixel_neighbors[i, j]
            regularization_matrix[i, i] += regularization_coefficient
            regularization_matrix[i, neighbor_index] -= regularization_coefficient

    return regularization_matrix

@numba.jit(nopython=True, parallel=True)
def weighted_pixel_signals_from_images(pixels, signal_scale, regular_to_pix, galaxy_image):
    """Compute the (scaled) signal in each pixel, where the signal is the sum of its datas_-pixel fluxes. \
    These pixel-signals are used to compute the effective regularization weight of each pixel.

    The pixel signals are scaled in the following ways:

    1) Divided by the number of datas_-pixels in the pixel, to ensure all pixels have the same \
    'relative' signal (i.e. a pixel with 10 regular-pixels doesn't have x2 the signal of one with 5).

    2) Divided by the maximum pixel-signal, so that all signals vary between 0 and 1. This ensures that the \
    regularizations weights are defined identically for any datas_ units or signal-to-noise ratio.

    3) Raised to the power of the hyper-parameter *signal_scale*, so the method can control the relative \
    contribution regularization in different regions of pixelization.
    """

    pixel_signals = np.zeros((pixels,))
    pixel_sizes = np.zeros((pixels,))

    for regular_index in range(galaxy_image.shape[0]):
        pixel_signals[regular_to_pix[regular_index]] += galaxy_image[regular_index]
        pixel_sizes[regular_to_pix[regular_index]] += 1

    pixel_signals /= pixel_sizes
    pixel_signals /= np.max(pixel_signals)

    return pixel_signals ** signal_scale

def weighted_regularization_weights_from_pixel_signals(coefficients, pixel_signals):
    """Compute the regularization weights, which are the effective regularization coefficient of every \
    pixel. They are computed using the (scaled) pixel-signal of each pixel.

    Two regularization coefficients are used, corresponding to the:

    1) (pixel_signals) - pixels with a high pixel-signal (i.e. where the signal is located in the pixelization).
    2) (1.0 - pixel_signals) - pixels with a low pixel-signal (i.e. where the signal is not located in the \
     pixelization).
    """
    return (coefficients[0] * pixel_signals + coefficients[1] * (1.0 - pixel_signals)) ** 2.0

@numba.jit(nopython=True, parallel=True)
def weighted_regularization_matrix_from_pixel_neighbors(regularization_weights, pixel_neighbors,
                                                        pixel_neighbors_size):
    """ From the pixel-neighbors, setup the regularization matrix using the weighted regularization scheme.

    Parameters
    ----------
    regularization_weights : list(float)
        The regularization_matrix weight of each pixel
    pixel_neighbors : [[]]
        A list of the neighbors of each pixel.
    """

    pixels = len(regularization_weights)

    regularization_matrix = np.zeros(shape=(pixels, pixels))

    regularization_weight = regularization_weights ** 2.0

    for i in range(pixels):
        for j in range(pixel_neighbors_size[i]):
            neighbor_index = pixel_neighbors[i, j]
            regularization_matrix[i, i] += regularization_weight[neighbor_index]
            regularization_matrix[neighbor_index, neighbor_index] += regularization_weight[neighbor_index]
            regularization_matrix[i, neighbor_index] -= regularization_weight[neighbor_index]
            regularization_matrix[neighbor_index, i] -= regularization_weight[neighbor_index]

    return regularization_matrix


import sys
import os
import re
import getdist
import math
import numpy as np
from astropy import cosmology

sys.path.append("../autolens/")

import galaxy
from profiles import light_profiles, mass_profiles


def weighted_avg_and_std(values, weights):
    """
    Return the weighted average and standard deviation.

    values, weights -- Numpy ndarrays with the same shape.
    """
    average = np.average(values, weights=weights)
    # Fast and numerically precise:
    variance = np.average((values-average)**2, weights=weights)
    return (average, math.sqrt(variance))

def getModelFolders(dir):
    regex = re.compile('.*Hyper')
    folders = [name for name in os.listdir(dir) if os.path.isdir(dir+name) and re.match(regex, name) == None]
    return folders

ltm_skip = 0
center_skip = 0

image_dir = '/gpfs/data_vector/pdtw24/PL_Data/SL03_2/'  # Dir of Object to make evidence tables from
#
# image_name = 'SLACSJ0252+0039'
# redshift = 0.2803
# source_redshift = 0.9818
# einstein_radius = 1.02
# source_light_min = 0.8
# source_light_max = 1.3
# slacs_ein_r_kpc = 4.18
# slacs_mass = '{0:e}'.format(10**11.25)
# slacs_radius = 4.4
# slacs_chab_stellar_mass =  '{0:e}'.format(10**11.21)
# slacs_chab_stellar_mass_lower =  '{0:e}'.format(10**11.08)
# slacs_chab_stellar_mass_upper =  '{0:e}'.format(10**11.34)
# slacs_sal_stellar_mass =  '{0:e}'.format(10**11.46)
# slacs_sal_stellar_mass_lower =  '{0:e}'.format(10**11.33)
# slacs_sal_stellar_mass_upper =  '{0:e}'.format(10**11.59)
#
image_name = 'SLACSJ1250+0523'
redshift = 0.2318
source_redshift = 0.7953
einstein_radius = 1.120
source_light_min = 0.5
source_light_max = 1.5
slacs_ein_r_kpc = 4.18
slacs_mass =  '{0:e}'.format(10**11.26)
slacs_chab_stellar_mass =  '{0:e}'.format(10**11.53)
slacs_chab_stellar_mass_lower =  '{0:e}'.format(10**11.46)
slacs_chab_stellar_mass_upper =  '{0:e}'.format(10**11.60)
slacs_sal_stellar_mass =  '{0:e}'.format(10**11.77)
slacs_sal_stellar_mass_lower =  '{0:e}'.format(10**11.7)
slacs_sal_stellar_mass_upper =  '{0:e}'.format(10**12.84)

# image_name = 'SLACSJ1430+4105'
# redshift = 0.2850
# source_redshift = 0.5753
# einstein_radius = 1.468
# source_light_min = 0.2
# source_light_max = 2.3
# our_einr_arc = 1.4668492852
# slacs_ein_r_kpc = 6.53
# slacs_mass =  '{0:e}'.format(10**11.73)
# slacs_chab_stellar_mass =  '{0:e}'.format(10**11.68)
# slacs_chab_stellar_mass_lower =  '{0:e}'.format(10**11.56)
# slacs_chab_stellar_mass_upper =  '{0:e}'.format(10**11.8)
# slacs_sal_stellar_mass =  '{0:e}'.format(10**11.93)
# slacs_sal_stellar_mass_lower =  '{0:e}'.format(10**11.82)
# slacs_sal_stellar_mass_upper =  '{0:e}'.format(10**12.04)

chab_stellar_frac = 0.33
chab_stellar_frac_error = 0.09

sal_stellar_frac = 0.59
sal_stellar_frac_error = 0.16

pipeline_folder = 'Pipeline_LMDM'
phase_folder = 'PL_Phase_6'

image_dir = image_dir + image_name + '/' + pipeline_folder + '/' + phase_folder + '/'

model_folder = 'SersicEll[XY_New][Rot_New]+ExpEll[XY_Prev][Rot_Prev]+NFWSph[XY_Fg][Rot_Off]+LTM[XY_Fg][Rot_Fg]+Shear'

pdf_file = image_dir + model_folder + '/' + image_name + '.txt'

pdf = getdist.mcsamples.loadMCSamples(pdf_file)
params = pdf.paramNames

sample_weights = []

total_mass_all = []
stellar_mass_all = []

for i in range(len(pdf.samples)):

    if pdf.weights[i] > 1e-6:

        values = pdf.samples[i]

    #    values = pdf.getMeans()

        sersic_bulge = mass_profiles.SersicMassProfile(centre=(values[0], values[1]), axis_ratio=values[5], phi=values[6], intensity=values[2], effective_radius=values[3],
                                                       sersic_index=values[4], mass_to_light_ratio=values[11])

        exponential_halo = mass_profiles.ExponentialMassProfile(axis_ratio=values[9], phi=values[6], intensity=values[7],
                                                                effective_radius=values[8], mass_to_light_ratio=values[11])

        dark_matter_halo = mass_profiles.SphericalNFWMassProfile(kappa_s=values[10], scale_radius=20.0)

        lens_galaxy=galaxy.Galaxy(redshift=redshift, mass_profiles=[sersic_bulge, exponential_halo, dark_matter_halo])

        lens_galaxy.einstein_radius = einstein_radius
        lens_galaxy.source_light_min = source_light_min
        lens_galaxy.source_light_max = source_light_max

        source_galaxy = galaxy.Galaxy(redshift=source_redshift)

        galaxy.LensingPlanes(galaxies=[lens_galaxy, source_galaxy], cosmological_model=cosmology.LambdaCDM(H0=70, Om0=0.3, Ode0=0.7))

        scale_r = 30*lens_galaxy.arcsec_per_kpc
        dark_matter_halo.scale_radius = scale_r
        lens_galaxy.mass_profiles[2].scale_radius = scale_r

        radius = slacs_ein_r_kpc*lens_galaxy.arcsec_per_kpc
        radius = 200.0*lens_galaxy.arcsec_per_kpc

        sample_weights.append(pdf.weights[i])

        total_mass_all.append(lens_galaxy.mass_within_circles(radius=radius))
        masses = lens_galaxy.mass_within_circles_individual(radius=radius)
        stellar_mass_all.append((masses[0] + masses[1]))

print(masses)

total_mass, total_mass_std = weighted_avg_and_std(total_mass_all, sample_weights)
total_mass_lower = total_mass - total_mass_std
total_mass_upper = total_mass + total_mass_std

stellar_mass, stellar_mass_std = weighted_avg_and_std(stellar_mass_all, sample_weights)
stellar_mass_lower = stellar_mass - stellar_mass_std
stellar_mass_upper = stellar_mass + stellar_mass_std

print()
print('Our Mass = ', '{0:e}'.format(total_mass), '(', '{0:e}'.format(total_mass_lower), '{0:e}'.format(total_mass_upper), ')')
print('SLACS Mass in R_Ein= ', slacs_mass)
print('Our Mass / SLACS Mass = ' , total_mass / float(slacs_mass) )

print()
print('Stellar Mass fraction in R_Ein = ', stellar_mass / total_mass, '(', stellar_mass_lower / total_mass, stellar_mass_upper / total_mass, ')' )
print('SLACS Chabrier Stellar Mass Fraction in R_Ein = ', chab_stellar_frac, ' +- ', chab_stellar_frac_error )
print('SLACS Salpeter Stellar Mass Fraction in R_Ein = ', sal_stellar_frac, ' +- ', sal_stellar_frac_error  )

print()
print('Our Stellar Mass = ','{0:e}'.format((stellar_mass)), '(', '{0:e}'.format(stellar_mass_lower), '{0:e}'.format(stellar_mass_upper), ')' )
print('SLACS Chabrier stellar mass = ', '{0:e}'.format(float(slacs_chab_stellar_mass)), '(', '{0:e}'.format(float(slacs_chab_stellar_mass_lower)), '{0:e}'.format(float(slacs_chab_stellar_mass_upper)), ')'  )
print('SLACS Salpeter stellar mass = ', '{0:e}'.format(float(slacs_sal_stellar_mass)), '(', '{0:e}'.format(float(slacs_sal_stellar_mass_lower)), '{0:e}'.format(float(slacs_sal_stellar_mass_upper)), ')'  )
print('Our Stellar Mass / SLACS Chabrier Stellar Mass = ', stellar_mass / float(slacs_chab_stellar_mass))
print('Our Stellar Mass / SLACS Salpeter Stellar Mass = ', stellar_mass / float(slacs_sal_stellar_mass))

lens_galaxy.plot_density_as_function_of_radius(maximum_radius=20.0*lens_galaxy.arcsec_per_kpc, number_bins=300, image_name=image_name, plot_errors=False,
                                               labels=['Sersic Bulge', 'EllipticalExponentialMass Stellar Halo', 'Dark Matter Halo'])


import getdist
import re
import os
import math
import numpy as np
from astropy import cosmology
import matplotlib.pyplot as plt

from autolens.analysis import galaxy
from autolens.model.profiles import mass_profiles

def weighted_avg_and_std(values, weights):
    """
    Return the weighted average and standard deviation.

    values, weights -- Numpy ndarrays with the same shape.
    """
    average = np.average(values, weights=weights)
    # Fast and numerically precise:
    variance = np.average((values-average)**2, weights=weights)
    return (average, math.sqrt(variance))

def getModelFolders(dir):
    regex = re.compile('.*Hyper')
    folders = [name for name in os.listdir(dir) if os.path.isdir(dir+name) and re.match(regex, name) == None]
    return folders

class SLACS(object):
            
    def __init__(self, image_dir, image_name):

        self.image_dir = image_dir
        self.image_name = image_name

        cosmological_model = cosmology.LambdaCDM(H0=70, Om0=0.3, Ode0=0.7)

        if image_name == 'SLACSJ0252+0039':

            self.redshift = 0.2803
            self.source_redshift = 0.9818
            self.einstein_radius = 1.02
            self.source_light_min = 0.8
            self.source_light_max = 1.3
            self.ein_r_kpc = 4.18
            self.mass = '{0:e}'.format(10 ** 11.25)
            self.radius = 4.4
            self.chab_stellar_mass = '{0:e}'.format(10 ** 11.21)
            self.chab_stellar_mass_lower = '{0:e}'.format(10 ** 11.08)
            self.chab_stellar_mass_upper = '{0:e}'.format(10 ** 11.34)
            self.sal_stellar_mass = '{0:e}'.format(10 ** 11.46)
            self.sal_stellar_mass_lower = '{0:e}'.format(10 ** 11.33)
            self.sal_stellar_mass_upper = '{0:e}'.format(10 ** 11.59)
            self.chab_stellar_frac = 0.4
            self.chab_stellar_frac_error = 0.12
            self.sal_stellar_frac = 0.71
            self.sal_stellar_frac_error = 0.21
            
        elif image_name == 'SLACSJ1250+0523':

            self.redshift = 0.2318
            self.source_redshift = 0.7953
            self.einstein_radius = 1.120
            self.source_light_min = 0.5
            self.source_light_max = 1.5
            self.ein_r_kpc = 4.18
            self.mass = '{0:e}'.format(10 ** 11.26)
            self.chab_stellar_mass = '{0:e}'.format(10 ** 11.53)
            self.chab_stellar_mass_lower = '{0:e}'.format(10 ** 11.46)
            self.chab_stellar_mass_upper = '{0:e}'.format(10 ** 11.60)
            self.sal_stellar_mass = '{0:e}'.format(10 ** 11.77)
            self.sal_stellar_mass_lower = '{0:e}'.format(10 ** 11.7)
            self.sal_stellar_mass_upper = '{0:e}'.format(10 ** 12.84)
            self.chab_stellar_frac = 0.68
            self.chab_stellar_frac_error = 0.11
            self.sal_stellar_frac = 1.2
            self.sal_stellar_frac_error = 0.19

        elif image_name == 'SLACSJ1430+4105':

            self.redshift = 0.2850
            self.source_redshift = 0.5753
            self.einstein_radius = 1.468
            self.source_light_min = 0.2
            self.source_light_max = 2.3
            self.our_einr_arc = 1.4668492852
            self.ein_r_kpc = 6.53
            self.mass = '{0:e}'.format(10 ** 11.73)
            self.chab_stellar_mass = '{0:e}'.format(10 ** 11.68)
            self.chab_stellar_mass_lower = '{0:e}'.format(10 ** 11.56)
            self.chab_stellar_mass_upper = '{0:e}'.format(10 ** 11.8)
            self.sal_stellar_mass = '{0:e}'.format(10 ** 11.93)
            self.sal_stellar_mass_lower = '{0:e}'.format(10 ** 11.82)
            self.sal_stellar_mass_upper = '{0:e}'.format(10 ** 12.04)
            self.chab_stellar_frac = 0.33
            self.chab_stellar_frac_error = 0.09
            self.sal_stellar_frac = 0.59
            self.sal_stellar_frac_error = 0.16

        self.arcsec_per_kpc = cosmological_model.arcsec_per_kpc_proper(self.redshift).value
        self.kpc_per_arcsec = 1.0/self.arcsec_per_kpc

        print(self.kpc_per_arcsec)
        # stop

    def load_samples(self, pdf_file, center_skip, ltm_skip):

        self.pdf = getdist.mcsamples.loadMCSamples(pdf_file)
        self.params = self.pdf.paramNames
        
        self.center_skip = center_skip
        self.ltm_skip = ltm_skip

    def setup_lensing_plane(self, values):

        sersic_bulge = mass_profiles.EllipticalSersicMass(centre=(values[0], values[1]), axis_ratio=values[5],
                                                          phi=values[6], intensity=values[2],
                                                          effective_radius=values[3],
                                                          sersic_index=values[4],
                                                          mass_to_light_ratio=values[12 + self.center_skip])

        exponential_halo = mass_profiles.EllipticalExponentialMass(axis_ratio=values[9 + self.center_skip],
                                                                   phi=values[10 + self.center_skip],
                                                                   intensity=values[7 + self.center_skip],
                                                                   effective_radius=values[8 + self.center_skip],
                                                                   mass_to_light_ratio=values[
                                                                    12 + self.ltm_skip + self.center_skip])

        dark_matter_halo = mass_profiles.SphericalNFW(kappa_s=values[11 + self.center_skip],
                                                      scale_radius=30.0*self.arcsec_per_kpc)

        self.lens_galaxy = galaxy.Galaxy(redshift=self.redshift,
                                         mass_profiles=[sersic_bulge, exponential_halo, dark_matter_halo])

        self.source_galaxy = galaxy.Galaxy(redshift=self.source_redshift)

    def setup_lensing_plane_ltm2r(self, values):

        sersic_bulge = mass_profiles.EllipticalSersicMass(centre=(values[0], values[1]), axis_ratio=values[5],
                                                          phi=values[6], intensity=values[2],
                                                          effective_radius=values[3],
                                                          sersic_index=values[4],
                                                          mass_to_light_ratio=values[12 + self.center_skip])

        exponential_halo = mass_profiles.EllipticalSersicMassRadialGradient(axis_ratio=values[9 + self.center_skip],
                                                                            phi=values[10 + self.center_skip],
                                                                            intensity=values[7 + self.center_skip],
                                                                            effective_radius=values[8 + self.center_skip],
                                                                            sersic_index=1.0,
                                                                            mass_to_light_ratio=values[
                                                                    12 + self.ltm_skip + self.center_skip],
                                                                            mass_to_light_gradient=values[
                                                                    13 + self.ltm_skip + self.center_skip])

        dark_matter_halo = mass_profiles.SphericalNFW(kappa_s=values[11 + self.center_skip],
                                                      scale_radius=30.0*self.arcsec_per_kpc)

        self.lens_galaxy = galaxy.Galaxy(redshift=self.redshift,
                                         mass_profiles=[sersic_bulge, exponential_halo, dark_matter_halo])

        self.source_galaxy = galaxy.Galaxy(redshift=self.source_redshift)

    def setup_model(self, model_index, ltmr2=False):

        values = self.pdf.samples[model_index]

    def masses_of_all_samples(self, radius_kpc):

        model_indexes = []
        sample_weights = []
        total_masses = []
        stellar_masses = []
        dark_masses = []
        stellar_fractions = []

        for i in range(len(self.pdf.samples)):

            if self.pdf.weights[i] > 1e-6:

                model_indexes.append(i)

                values = self.pdf.samples[i]

                self.setup_lensing_plane(values)

                sample_weights.append(self.pdf.weights[i])

                radius_arc = radius_kpc * self.arcsec_per_kpc
                masses = self.lens_galaxy.dimensionless_mass_within_circles_individual(radius=radius_arc)

                total_mass = sum(masses)
                stellar_mass = (masses[0] + masses[1])
                dark_mass = masses[2]

                total_masses.append(total_mass)
                stellar_masses.append(stellar_mass)
                dark_masses.append(dark_mass)
                stellar_fractions.append(stellar_mass/total_mass)

        return model_indexes, sample_weights, total_masses, stellar_masses, dark_masses, stellar_fractions

    def weighted_densities_vs_radii(self, radius_kpc, weight_cut=1e-6, number_bins=60):

        values = self.pdf.samples[1]
        self.setup_lensing_plane(values)
        radii = list(np.linspace(5e-3, radius_kpc*self.arcsec_per_kpc, number_bins+1))

        self.bulge_density_plot = []
        self.bulge_density_upper_plot = []
        self.bulge_density_lower_plot = []

        self.halo_density_plot = []
        self.halo_density_upper_plot = []
        self.halo_density_lower_plot = []

        self.dark_density_plot = []
        self.dark_density_upper_plot = []
        self.dark_density_lower_plot = []

        self.radii_plot = []

        for r in range(number_bins):

            annuli_area = (math.pi * radii[r + 1] ** 2 - math.pi * radii[r] ** 2)

            weights = []
            bulge_densities = []
            halo_densities = []
            dark_densities = []

            for i in range(len(self.pdf.samples)):

                if self.pdf.weights[i] > weight_cut:

                    weights.append(self.pdf.weights[i])
                    values = self.pdf.samples[i]
                    self.setup_lensing_plane(values)

                    densities = ((self.lens_galaxy.dimensionless_mass_within_circles_individual(radii[r + 1]) -
                                  self.lens_galaxy.dimensionless_mass_within_circles_individual(radii[r])) / annuli_area)

                    bulge_densities.append(densities[0])
                    halo_densities.append(densities[1])
                    dark_densities.append(densities[2])


            bulge_avg, bulge_std = weighted_avg_and_std(np.asarray(bulge_densities), np.asarray(weights))
            halo_avg, halo_std = weighted_avg_and_std(np.asarray(halo_densities), np.asarray(weights))
            dark_avg, dark_std = weighted_avg_and_std(np.asarray(dark_densities), np.asarray(weights))

            self.bulge_density_plot.append(bulge_avg)
            self.bulge_density_upper_plot.append(bulge_avg + bulge_std)
            self.bulge_density_lower_plot.append(bulge_avg - bulge_std)

            self.halo_density_plot.append(halo_avg)
            self.halo_density_upper_plot.append(halo_avg + halo_std)
            self.halo_density_lower_plot.append(halo_avg - halo_std)

            self.dark_density_plot.append(dark_avg)
            self.dark_density_upper_plot.append(dark_avg + dark_std)
            self.dark_density_lower_plot.append(dark_avg - dark_std)

            self.radii_plot.append( ((radii[r+1] + radii[r]) / 2.0) * self.kpc_per_arcsec)

    def plot_density(self, image_name='', labels=None, xaxis_is_physical=True, yaxis_is_physical=True):

        plt.title('Decomposed surface density profile of ' + image_name, size=20)

        if xaxis_is_physical:
            plt.xlabel('Distance From Galaxy Center (kpc)', size=20)
        else:
            plt.xlabel('Distance From Galaxy Center (")', size=20)

        if yaxis_is_physical:
            plt.ylabel(r'Surface Mass Density $\Sigma$ ($M_{\odot}$ / kpc $^2$)', size=20)
        else:
            pass

        plt.semilogy(self.radii_plot, self.bulge_density_plot, color='r', label='Sersic Bulge')
        plt.semilogy(self.radii_plot, self.halo_density_plot, color='g', label='EllipticalExponentialMass Halo')
        plt.semilogy(self.radii_plot, self.dark_density_plot, color='k', label='Dark Matter Halo')

        plt.semilogy(self.radii_plot, self.bulge_density_upper_plot, color='r', linestyle='--')
        plt.semilogy(self.radii_plot, self.halo_density_upper_plot, color='g', linestyle='--')
        plt.semilogy(self.radii_plot, self.dark_density_upper_plot, color='k', linestyle='--')

        plt.semilogy(self.radii_plot, self.bulge_density_lower_plot, color='r', linestyle='--')
        plt.semilogy(self.radii_plot, self.halo_density_lower_plot, color='g', linestyle='--')
        plt.semilogy(self.radii_plot, self.dark_density_lower_plot, color='k', linestyle='--')

        # plt.loglog(radii_plot, density_0, color='r', label='Sersic Bulge')
        # plt.loglog(radii_plot, density_1, color='g', label='EllipticalExponentialMass Halo')
        # plt.loglog(radii_plot, density_2, color='k', label='Dark Matter Halo')

        # plt.semilogy(radii_plot, density_lower[:, 0], color='r', linestyle='--')
        # plt.semilogy(radii_plot, density_upper[:, 0], color='r', linestyle='--')
        # plt.semilogy(radii_plot, density_lower[:, 1], color='g', linestyle='--')
        # plt.semilogy(radii_plot, density_upper[:, 1], color='g', linestyle='--')
        # plt.semilogy(radii_plot, density_lower[:, 2], color='k', linestyle='--')
        # plt.semilogy(radii_plot, density_upper[:, 2], color='k', linestyle='--')

        plt.axvline(x=self.einstein_radius*self.kpc_per_arcsec, linestyle='--')
        plt.axvline(x=self.source_light_min*self.kpc_per_arcsec, linestyle='-')
        plt.axvline(x=self.source_light_max*self.kpc_per_arcsec, linestyle='-')

        plt.legend()
        plt.show()


import os
import shutil

import pytest

from autofit import conf
from autolens.model.galaxy import galaxy as g
from autolens.data.array import grids
from autolens.lensing import ray_tracing
from autolens.lensing.plotters import ray_tracing_plotters
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp


@pytest.fixture(name='general_config')
def test_general_config():
    general_config_path = "{}/../../test_files/configs/plotting/".format(os.path.dirname(os.path.realpath(__file__)))
    conf.instance.general = conf.NamedConfig(general_config_path + "general.ini")


@pytest.fixture(name='ray_tracing_plotter_path')
def test_ray_tracing_plotter_setup():
    galaxy_plotter_path = "{}/../../test_files/plotting/ray_tracing/".format(os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(galaxy_plotter_path):
        shutil.rmtree(galaxy_plotter_path)

    os.mkdir(galaxy_plotter_path)

    return galaxy_plotter_path


@pytest.fixture(name='galaxy_light')
def test_galaxy_light():
    return g.Galaxy(light=lp.EllipticalSersic(intensity=1.0))


@pytest.fixture(name='galaxy_mass')
def test_galaxy_mass():
    return g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=1.0))


@pytest.fixture(name='grids')
def test_grids():
    return grids.DataGrids.from_shape_and_pixel_scale(shape=(100, 100), pixel_scale=0.05, sub_grid_size=2)


@pytest.fixture(name='tracer')
def test_tracer(galaxy_light, galaxy_mass, grids):
    return ray_tracing.TracerImageSourcePlanes(lens_galaxies=[galaxy_mass, galaxy_light],
                                               source_galaxies=[galaxy_light],
                                               image_plane_grids=[grids])


def test__tracer_sub_plot_output_dependent_on_config(tracer, general_config, ray_tracing_plotter_path):
    ray_tracing_plotters.plot_ray_tracing_subplot(tracer=tracer, output_path=ray_tracing_plotter_path,
                                                  output_format='png')

    assert os.path.isfile(path=ray_tracing_plotter_path + 'tracer_normal.png')
    os.remove(path=ray_tracing_plotter_path + 'tracer_normal.png')


def test__tracer_individuals__depedent_on_config(tracer, general_config, ray_tracing_plotter_path):
    ray_tracing_plotters.plot_ray_tracing_individual(tracer=tracer, output_path=ray_tracing_plotter_path,
                                                     output_format='png')

    assert os.path.isfile(path=ray_tracing_plotter_path + 'tracer_image_plane_image.png')
    os.remove(path=ray_tracing_plotter_path + 'tracer_image_plane_image.png')

    assert os.path.isfile(path=ray_tracing_plotter_path + 'tracer_source_plane.png')
    os.remove(path=ray_tracing_plotter_path + 'tracer_source_plane.png')

    assert not os.path.isfile(path=ray_tracing_plotter_path + 'tracer_surface_density.png')

    assert os.path.isfile(path=ray_tracing_plotter_path + 'tracer_potential.png')
    os.remove(path=ray_tracing_plotter_path + 'tracer_potential.png')

    assert not os.path.isfile(path=ray_tracing_plotter_path + 'tracer_deflections_y.png')
    assert not os.path.isfile(path=ray_tracing_plotter_path + 'tracer_deflections_x.png')


def test__image_plane_image_is_output(tracer, ray_tracing_plotter_path):
    ray_tracing_plotters.plot_image_plane_image(tracer=tracer, output_path=ray_tracing_plotter_path,
                                                output_format='png')
    assert os.path.isfile(path=ray_tracing_plotter_path + 'tracer_image_plane_image.png')
    os.remove(path=ray_tracing_plotter_path + 'tracer_image_plane_image.png')


def test__surface_density_is_output(tracer, ray_tracing_plotter_path):
    ray_tracing_plotters.plot_surface_density(tracer=tracer, output_path=ray_tracing_plotter_path, output_format='png')
    assert os.path.isfile(path=ray_tracing_plotter_path + 'tracer_surface_density.png')
    os.remove(path=ray_tracing_plotter_path + 'tracer_surface_density.png')


def test__potential_is_output(tracer, ray_tracing_plotter_path):
    ray_tracing_plotters.plot_potential(tracer=tracer, output_path=ray_tracing_plotter_path, output_format='png')
    assert os.path.isfile(path=ray_tracing_plotter_path + 'tracer_potential.png')
    os.remove(path=ray_tracing_plotter_path + 'tracer_potential.png')


def test__deflections_y_is_output(tracer, ray_tracing_plotter_path):
    ray_tracing_plotters.plot_deflections_y(tracer=tracer, output_path=ray_tracing_plotter_path, output_format='png')
    assert os.path.isfile(path=ray_tracing_plotter_path + 'tracer_deflections_y.png')
    os.remove(path=ray_tracing_plotter_path + 'tracer_deflections_y.png')


def test__deflections_x_is_output(tracer, ray_tracing_plotter_path):
    ray_tracing_plotters.plot_deflections_x(tracer=tracer, output_path=ray_tracing_plotter_path, output_format='png')
    assert os.path.isfile(path=ray_tracing_plotter_path + 'tracer_deflections_x.png')
    os.remove(path=ray_tracing_plotter_path + 'tracer_deflections_x.png')



import numpy as np

from autolens.model.inversion.util import regularization_util as reg_util


class TestRegularizationConstantMatrix:

    def test__1_b_matrix_size_3x3__weights_all_1s__makes_correct_regularization_matrix(self):
        # Here, we define the pixel_neighbors first here and make the B matrices based on them.

        # You'll notice that actually, the B Matrix doesn't have to have the -1's going down the diagonal and we
        # don't have to have as many B matrices as we do the pix pixel with the most  vertices. We can combine
        # the rows of each B matrix wherever we like ;0.

        pixel_neighbors = np.array([[1, 2, -1],
                                    [0, -1, -1],
                                    [0, -1, -1]])

        pixel_neighbors_size = np.array([2, 1, 1])

        test_b_matrix = np.array([[-1, 1, 0],  # Pair 1
                                  [-1, 0, 1],  # Pair 2
                                  [0, 0, 0, ]])  # Pair 1 flip

        test_regularization_matrix = np.matmul(test_b_matrix.T, test_b_matrix) + 1e-8 * np.identity(3)

        regularization_matrix = reg_util.constant_regularization_matrix_from_pixel_neighbors(coefficients=(1.0,),
            pixel_neighbors=pixel_neighbors, pixel_neighbors_size=pixel_neighbors_size)

        assert (regularization_matrix == test_regularization_matrix).all()
        assert (abs(np.linalg.det(regularization_matrix)) > 1e-8)

    def test__1_b_matrix_size_4x4__weights_all_1s__makes_correct_regularization_matrix(self):

        test_b_matrix = np.array([[-1, 1, 0, 0],
                                  [0, -1, 1, 0],
                                  [0, 0, -1, 1],
                                  [1, 0, 0, -1]])

        test_regularization_matrix = np.matmul(test_b_matrix.T, test_b_matrix) + 1e-8 * np.identity(4)

        pixel_neighbors = np.array([[1, 3, -1, -1],
                                    [0, 2, -1, -1],
                                    [1, 3, -1, -1],
                                    [0, 2, -1, -1]])

        pixel_neighbors_size = np.array([2, 2, 2, 2])

        regularization_matrix = reg_util.constant_regularization_matrix_from_pixel_neighbors(coefficients=(1.0,),
            pixel_neighbors=pixel_neighbors, pixel_neighbors_size=pixel_neighbors_size)

        assert (regularization_matrix == test_regularization_matrix).all()
        assert (abs(np.linalg.det(regularization_matrix)) > 1e-8)

    def test__1_b_matrix_size_4x4__coefficient_2__makes_correct_regularization_matrix(self):

        pixel_neighbors = np.array([[1, 3, -1, -1],
                                    [0, 2, -1, -1],
                                    [1, 3, -1, -1],
                                    [0, 2, -1, -1]])

        pixel_neighbors_size = np.array([2, 2, 2, 2])

        test_b_matrix = 2.0 * np.array([[-1, 1, 0, 0],
                                        [0, -1, 1, 0],
                                        [0, 0, -1, 1],
                                        [1, 0, 0, -1]])

        test_regularization_matrix = np.matmul(test_b_matrix.T, test_b_matrix) + 1e-8 * np.identity(4)

        regularization_matrix = reg_util.constant_regularization_matrix_from_pixel_neighbors(coefficients=(2.0,),
            pixel_neighbors=pixel_neighbors, pixel_neighbors_size=pixel_neighbors_size)

        assert (regularization_matrix == test_regularization_matrix).all()
        assert (abs(np.linalg.det(regularization_matrix)) > 1e-8)

    def test__1_b_matrix_size_9x9__coefficient_2__makes_correct_regularization_matrix(self):

        pixel_neighbors = np.array([[1, 3, -1, -1],
                                   [4, 2, 0, -1],
                                   [1, 5, -1, -1],
                                   [4, 6, 0, -1],
                                   [7, 1, 5, 3],
                                   [4, 2, 8, -1],
                                   [7, 3, -1, -1],
                                   [4, 8, 6, -1],
                                   [7, 5, -1, -1]])

        pixel_neighbors_size = np.array([2, 3, 2, 3, 4, 3, 2, 3, 2])

        test_b_matrix_0 = np.array([[-1, 1, 0, 0, 0, 0, 0, 0, 0],
                                    [-1, 0, 0, 1, 0, 0, 0, 0, 0],
                                    [0, -1, 1, 0, 0, 0, 0, 0, 0],
                                    [0, -1, 0, 0, 1, 0, 0, 0, 0],
                                    [0, 0, -1, 0, 0, 1, 0, 0, 0],
                                    [0, 0, 0, -1, 1, 0, 0, 0, 0],
                                    [0, 0, 0, -1, 0, 0, 1, 0, 0],
                                    [0, 0, 0, 0, -1, 1, 0, 0, 0],
                                    [0, 0, 0, 0, -1, 0, 0, 1, 0]])

        test_b_matrix_1 = np.array([[0, 0, 0, 0, 0, -1, 0, 0, 1],
                                    [0, 0, 0, 0, 0, 0, -1, 1, 0],
                                    [0, 0, 0, 0, 0, 0, 0, -1, 1],
                                    [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0, 0, 0, 0]])

        test_regularization_matrix_0 = np.matmul(test_b_matrix_0.T, test_b_matrix_0)
        test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)

        test_regularization_matrix = test_regularization_matrix_0 + \
                                     test_regularization_matrix_1 + 1e-8 * np.identity(9)

        regularization_matrix = reg_util.constant_regularization_matrix_from_pixel_neighbors(coefficients=(1.0,),
            pixel_neighbors=pixel_neighbors, pixel_neighbors_size=pixel_neighbors_size)

        assert (regularization_matrix == test_regularization_matrix).all()
        assert (abs(np.linalg.det(regularization_matrix)) > 1e-8)
        
        
class TestRegularizationWeightedPixelSignals:

    def test__x3_image_pixels_signals_1s__pixel_scale_1__pixel_signals_all_1s(self):

        regular_to_pix = np.array([0, 1, 2])
        galaxy_image = np.array([1.0, 1.0, 1.0])

        pixel_signals = reg_util.weighted_pixel_signals_from_images(pixels=3, signal_scale=1.0,  
                                                                    regular_to_pix=regular_to_pix,
                                                                    galaxy_image=galaxy_image)

        assert (pixel_signals == np.array([1.0, 1.0, 1.0])).all()

    def test__x4_image_pixels_signals_1s__pixel_signals_still_all_1s(self):

        regular_to_pix = np.array([0, 1, 2, 0])
        galaxy_image = np.array([1.0, 1.0, 1.0, 1.0])

        pixel_signals = reg_util.weighted_pixel_signals_from_images(pixels=3, signal_scale=1.0,
                                                                    regular_to_pix=regular_to_pix,
                                                                    galaxy_image=galaxy_image)

        assert (pixel_signals == np.array([1.0, 1.0, 1.0])).all()

    def test__galaxy_flux_in_a_pixel_pixel_is_double_the_others__pixel_signal_is_1_others_a_half(self):

        regular_to_pix = np.array([0, 1, 2])
        galaxy_image = np.array([2.0, 1.0, 1.0])

        pixel_signals = reg_util.weighted_pixel_signals_from_images(pixels=3, signal_scale=1.0,
                                                                    regular_to_pix=regular_to_pix,
                                                                    galaxy_image=galaxy_image)

        assert (pixel_signals == np.array([1.0, 0.5, 0.5])).all()

    def test__same_as_above_but_pixel_scale_2__scales_pixel_signals(self):

        regular_to_pix = np.array([0, 1, 2])
        galaxy_image = np.array([2.0, 1.0, 1.0])

        pixel_signals = reg_util.weighted_pixel_signals_from_images(pixels=3, signal_scale=2.0,
                                                                    regular_to_pix=regular_to_pix,
                                                                    galaxy_image=galaxy_image)

        assert (pixel_signals == np.array([1.0, 0.25, 0.25])).all()


class TestRegularizationWeightedRegularizationWeights(object):

    def test__pixel_signals_all_1s__coefficients_all_1s__weights_all_1s(self):

        pixel_signals = np.array([1.0, 1.0, 1.0])

        weights = reg_util.weighted_regularization_weights_from_pixel_signals(coefficients=(1.0, 1.0),
                                                                              pixel_signals=pixel_signals)

        assert (weights == np.array([1.0, 1.0, 1.0])).all()

    def test__pixel_signals_vary__coefficents_all_1s__weights_still_all_1s(self):

        pixel_signals = np.array([0.25, 0.5, 0.75])

        weights = reg_util.weighted_regularization_weights_from_pixel_signals(coefficients=(1.0, 1.0),
                                                                              pixel_signals=pixel_signals)

        assert (weights == np.array([1.0, 1.0, 1.0])).all()

    def test__pixel_signals_vary__coefficents_1_and_0__weights_are_pixel_signals_squared(self):

        pixel_signals = np.array([0.25, 0.5, 0.75])

        weights = reg_util.weighted_regularization_weights_from_pixel_signals(coefficients=(1.0, 0.0),
                                                                              pixel_signals=pixel_signals)

        assert (weights == np.array([0.25 ** 2.0, 0.5 ** 2.0, 0.75 ** 2.0])).all()

    def test__pixel_signals_vary__coefficents_0_and_1__weights_are_1_minus_pixel_signals_squared(self):

        pixel_signals = np.array([0.25, 0.5, 0.75])

        weights = reg_util.weighted_regularization_weights_from_pixel_signals(coefficients=(0.0, 1.0),
                                                                              pixel_signals=pixel_signals)

        assert (weights == np.array([0.75 ** 2.0, 0.5 ** 2.0, 0.25 ** 2.0])).all()


class TestRegularizationWeightedMatrix:

    def test__1_b_matrix_size_4x4__weights_all_1s__makes_correct_regularization_matrix(self):

        pixel_neighbors = np.array([[2],
                                    [3],
                                    [0],
                                    [1]])

        pixel_neighbors_size = np.array([1, 1, 1, 1])

        test_b_matrix = np.array([[-1, 0, 1, 0],
                                  [0, -1, 0, 1],
                                  [1, 0, -1, 0],
                                  [0, 1, 0, -1]])

        test_regularization_matrix = np.matmul(test_b_matrix.T, test_b_matrix)

        regularization_weights = np.ones((4,))

        regularization_matrix = reg_util.weighted_regularization_matrix_from_pixel_neighbors(regularization_weights,
                                                                               pixel_neighbors, pixel_neighbors_size)

        assert (regularization_matrix == test_regularization_matrix).all()

    def test__2_b_matrices_size_3x3__weights_all_1s__makes_correct_regularization_matrix(self):
        # Here, we define the pixel_neighbors first here and make the B matrices based on them.

        # You'll notice that actually, the B Matrix doesn't have to have the -1's going down the diagonal and we
        # don't have to have as many B matrices as we do the pix pixel with the most  vertices. We can combine
        # the rows of each B matrix wherever we like ;0.

        pixel_neighbors = np.array([[1, 2],
                                    [0, -1],
                                    [0, -1]])

        pixel_neighbors_size = np.array([2, 1, 1])

        test_b_matrix_1 = np.array([[-1, 1, 0],  # Pair 1
                                    [-1, 0, 1],  # Pair 2
                                    [1, -1, 0]])  # Pair 1 flip

        test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)

        test_b_matrix_2 = np.array([[1, 0, -1],  # Pair 2 flip
                                    [0, 0, 0],
                                    [0, 0, 0]])

        test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)

        test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2

        regularization_weights = np.ones((3))

        regularization_matrix = reg_util.weighted_regularization_matrix_from_pixel_neighbors(regularization_weights,
                                                                               pixel_neighbors,
                                                                               pixel_neighbors_size)

        assert (regularization_matrix == test_regularization_matrix).all()

    def test__2_b_matrices_size_4x4__weights_all_1s__makes_correct_regularization_matrix(self):

        test_b_matrix_1 = np.array([[-1, 1, 0, 0],
                                    [0, -1, 1, 0],
                                    [0, 0, -1, 1],
                                    [1, 0, 0, -1]])

        test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)

        test_b_matrix_2 = np.array([[-1, 0, 0, 1],
                                    [1, -1, 0, 0],
                                    [0, 1, -1, 0],
                                    [0, 0, 1, -1]])

        test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)

        test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2

        pixel_neighbors = np.array([[1, 3],
                                    [0, 2],
                                    [1, 3],
                                    [0, 2]])

        pixel_neighbors_size = np.array([2, 2, 2, 2])

        regularization_weights = np.ones((4,))

        regularization_matrix = reg_util.weighted_regularization_matrix_from_pixel_neighbors(regularization_weights,
                                                                               pixel_neighbors, pixel_neighbors_size)

        assert (regularization_matrix == test_regularization_matrix).all()

    def test__4_b_matrices_size_6x6__weights_all_1s__makes_correct_regularization_matrix(self):
        # Again, lets exploit the freedom we have when setting up our B matrices to make matching it to pairs a
        # lot less Stressful.

        pixel_neighbors = np.array([[2, 3, 4, -1],
                                   [2, 5, -1, -1],
                                   [0, 1, 3, 5],
                                   [0, 2, -1, -1],
                                   [5, 0, -1, -1],
                                   [4, 1, 2, -1]])

        pixel_neighbors_size = np.array([3, 2, 4, 2, 2, 3])

        test_b_matrix_1 = np.array([[-1, 0, 1, 0, 0, 0],  # Pair 1
                                    [0, -1, 1, 0, 0, 0],  # Pair 2
                                    [-1, 0, 0, 1, 0, 0],  # Pair 3
                                    [0, 0, 0, 0, -1, 1],  # Pair 4
                                    [0, -1, 0, 0, 0, 1],  # Pair 5
                                    [-1, 0, 0, 0, 1, 0]])  # Pair 6

        test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)

        test_b_matrix_2 = np.array([[0, 0, -1, 1, 0, 0],  # Pair 7
                                    [0, 0, -1, 0, 0, 1],  # Pair 8
                                    [0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0]])

        test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)

        test_b_matrix_3 = np.array([[1, 0, -1, 0, 0, 0],  # Pair 1 flip
                                    [0, 1, -1, 0, 0, 0],  # Pair 2 flip
                                    [1, 0, 0, -1, 0, 0],  # Pair 3 flip
                                    [0, 0, 0, 0, 1, -1],  # Pair 4 flip
                                    [0, 1, 0, 0, 0, -1],  # Pair 5 flip
                                    [1, 0, 0, 0, -1, 0]])  # Pair 6 flip

        test_regularization_matrix_3 = np.matmul(test_b_matrix_3.T, test_b_matrix_3)

        test_b_matrix_4 = np.array([[0, 0, 1, -1, 0, 0],  # Pair 7 flip
                                    [0, 0, 1, 0, 0, -1],  # Pair 8 flip
                                    [0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0]])

        test_regularization_matrix_4 = np.matmul(test_b_matrix_4.T, test_b_matrix_4)

        test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2 + \
                                     test_regularization_matrix_3 + + test_regularization_matrix_4

        regularization_weights = np.ones((6))

        regularization_matrix = reg_util.weighted_regularization_matrix_from_pixel_neighbors(regularization_weights,
                                                                               pixel_neighbors, pixel_neighbors_size)

        assert (regularization_matrix == test_regularization_matrix).all()

    def test__2_b_matrices_size_4x4_variables_regularization_weights__makes_correct_regularization_matrix(self):
        # Simple case, where we have just one regularization direction, regularizing pixel 0 -> 1 and 1 -> 2.

        # This means our B matrix is:

        # [-1, 1, 0]
        # [0, -1, 1]
        # [0, 0, -1]

        # Regularization Matrix, H = B * B.T.I can

        regularization_weights = np.array([2.0, 4.0, 1.0, 8.0])

        test_b_matrix_1 = np.array([[-2, 2, 0, 0],
                                    [-2, 0, 2, 0],
                                    [0, -4, 4, 0],
                                    [0, -4, 0, 4]])

        test_b_matrix_2 = np.array([[4, -4, 0, 0],
                                    [1, 0, -1, 0],
                                    [0, 1, -1, 0],
                                    [0, 8, 0, -8]])

        test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)
        test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)

        test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2

        pixel_neighbors = np.array([[1, 2, -1, -1],
                                    [0, 2, 3, -1],
                                    [0, 1, -1, -1],
                                    [1, -1, -1, -1]])

        pixel_neighbors_size = np.array([2, 3, 2, 1])

        regularization_matrix = reg_util.weighted_regularization_matrix_from_pixel_neighbors(regularization_weights,
                                                                               pixel_neighbors, pixel_neighbors_size)

        assert (regularization_matrix == test_regularization_matrix).all()

    def test__4_b_matrices_size_6x6_with_regularization_weights__makes_correct_regularization_matrix(self):

        pixel_neighbors = np.array([[1, 4, -1, -1],
                                    [2, 4, 0, -1],
                                    [3, 4, 5, 1],
                                    [5, 2, -1, -1],
                                    [5, 0, 1, 2],
                                    [2, 3, 4, -1]])

        pixel_neighbors_size = np.array([2, 3, 4, 2, 4, 3])
        regularization_weights = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])

        # I'm inputting the regularization weights directly thiss time, as it'd be a pain to multiply with a
        # loop.

        test_b_matrix_1 = np.array([[-1, 1, 0, 0, 0, 0],  # Pair 1
                                    [-1, 0, 0, 0, 1, 0],  # Pair 2
                                    [0, -2, 2, 0, 0, 0],  # Pair 3
                                    [0, -2, 0, 0, 2, 0],  # Pair 4
                                    [0, 0, -3, 3, 0, 0],  # Pair 5
                                    [0, 0, -3, 0, 3, 0]])  # Pair 6

        test_b_matrix_2 = np.array([[0, 0, -3, 0, 0, 3],  # Pair 7
                                    [0, 0, 0, -4, 0, 4],  # Pair 8
                                    [0, 0, 0, 0, -5, 5],  # Pair 9
                                    [0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0]])

        # Now do the same pairs but with the regularization direction and weights swapped.

        test_b_matrix_3 = np.array([[2, -2, 0, 0, 0, 0],  # Pair 1
                                    [5, 0, 0, 0, -5, 0],  # Pair 2
                                    [0, 3, -3, 0, 0, 0],  # Pair 3
                                    [0, 5, 0, 0, -5, 0],  # Pair 4
                                    [0, 0, 4, -4, 0, 0],  # Pair 5
                                    [0, 0, 5, 0, -5, 0]])  # Pair 6

        test_b_matrix_4 = np.array([[0, 0, 6, 0, 0, -6],  # Pair 7
                                    [0, 0, 0, 6, 0, -6],  # Pair 8
                                    [0, 0, 0, 0, 6, -6],  # Pair 9
                                    [0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 0]])

        test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)
        test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)
        test_regularization_matrix_3 = np.matmul(test_b_matrix_3.T, test_b_matrix_3)
        test_regularization_matrix_4 = np.matmul(test_b_matrix_4.T, test_b_matrix_4)

        test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2 + \
                                     test_regularization_matrix_3 + test_regularization_matrix_4

        regularization_matrix = reg_util.weighted_regularization_matrix_from_pixel_neighbors(regularization_weights,
                                                                               pixel_neighbors, pixel_neighbors_size)

        assert (regularization_matrix == test_regularization_matrix).all()


import os
import shutil

from autofit.core import non_linear as nl
from autolens.model.galaxy import galaxy, galaxy_model as gm
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp
from test.integration import tools

dirpath = os.path.dirname(os.path.realpath(__file__))
dirpath = os.path.dirname(dirpath)
output_path = '{}/../output/model_mapper'.format(dirpath)


try:
    shutil.rmtree(output_path)
except FileNotFoundError:
    pass



def pipeline():

    pipeline_name = "link_variable_to_constant_floats"
    data_name = '/link_variable_to_constant_floats'

    tools.reset_paths(data_name, pipeline_name, output_path)

    sersic = lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, intensity=1.0, effective_radius=1.3,
                                 sersic_index=3.0)

    lens_galaxy = galaxy.Galaxy(light_profile=sersic)

    tools.simulate_integration_image(data_name=data_name, pixel_scale=0.5, lens_galaxies=[lens_galaxy],
                                     source_galaxies=[], target_signal_to_noise=10.0)
    image = tools.load_image(data_name=data_name, pixel_scale=0.5)

    pipeline = make_pipeline(pipeline_name=pipeline_name)

    results = pipeline.run(image=image)
    for result in results:
        print(result)


def make_pipeline(pipeline_name):
    class MMPhase(ph.LensPlanePhase):

        def pass_priors(self, previous_results):
            self.lens_galaxies[0].sersic.axis_ratio = 0.2
            self.lens_galaxies[0].sersic.phi = 90.0

    phase1 = MMPhase(lens_galaxies=[gm.GalaxyModel(sersic=lp.EllipticalSersic)],
                     optimizer_class=nl.MultiNest, phase_name="{}/phase1".format(pipeline_name))

    phase1.optimizer.n_live_points = 20
    phase1.optimizer.sampling_efficiency = 0.8

    class MMPhase2(ph.LensPlanePhase):

        def pass_priors(self, previous_results):
            self.lens_galaxies = previous_results[0].constant.lens_galaxies

    phase2 = MMPhase2(lens_galaxies=[gm.GalaxyModel(sersic=lp.EllipticalSersic)],
                      optimizer_class=nl.MultiNest, phase_name="{}/phase2".format(pipeline_name))

    phase2.optimizer.n_live_points = 20
    phase2.optimizer.sampling_efficiency = 0.8

    return pl.PipelineImaging(pipeline_name, phase1, phase2)


if __name__ == "__main__":
    pipeline()
